
import itertools
import pycuda.driver as cuda
import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype, context_dependent_memoize
# import pycuda.elementwise as elementwise
from pycuda.compiler import SourceModule

import numpy as np
import networkx as nx
import h5py
from datetime import datetime

from neurokernel.LPU.LPU import LPU

class BaseInputProcessor(object):
    def __init__(self, var_list, mode = 0, memory_mode = 'cpu',
                 input_file = None, input_interval = 1,
                 sensory_file = None, sensory_interval = 1):
        # var_list should be a list of (variable, uids)
        # If no uids is provided, the variable will be ignored
        # Invalid uids will be ignored
        # Derived classes should update self.variables[var]['input']
        # for each variable in update_input method with a ndarray of
        # length len(uids) and the correct dtype
        self.variables = {var:{'uids':uids,'input':None}
                          for var, uids in var_list if uids is not None}
        self.epoch = 0
        self.dest_inds = {}
        self._LPU_obj = None
        # mode = 0 => provide zero input when no input is available
        # mode = 1 => persist with previous input if no input is available
        self.mode = mode
        self.memory_mode = memory_mode
        assert memory_mode in ['cpu', 'gpu'], "Memory mode must be either 'cpu' or 'gpu'"
        self.input_to_be_processed = True
        self.dtypes = {}
        self._d_input = {}
        self.dest_inds = {}
        self.sensory_file = sensory_file
        self.sensory_interval = sensory_interval
        self.input_file = input_file
        self.input_interval = input_interval
        self.metadata = {}
        self._additional_graph = nx.MultiDiGraph()

    def get_new_graph(self):
        return self._additional_graph

    def add_variables(self, var_list):
        for var, uids in var_list:
            if len(uids):
                if var in self.variables:
                    self.variables[var]['uids'].append(uids)
                else:
                    self.variables[var] = {'uids':uids,'input':None}
        self.variables.update({var:{'uids':uids,'input':None}
                                for var, uids in var_list if len(uids)})

    @property
    def LPU_obj(self):
        return self._LPU_obj

    @LPU_obj.setter
    def LPU_obj(self, value):
        assert(isinstance(value, LPU))
        self._LPU_obj = value
        self.dt = self._LPU_obj.dt
        self.sim_dt = self._LPU_obj.dt
        self.memory_manager = self._LPU_obj.memory_manager

    def run_step(self):
        if not self.is_input_available():
            if self.mode == 0:
                self.input_to_be_processed = False
                # for var in self.variables:
                #     self._d_input[var].fill(0)
            elif self.mode == 1:
                self.input_to_be_processed = True
                pass
            else:
                self.input_to_be_processed = False
                # for var in self.variables:
                #     self._d_input[var].fill(0)
                self.LPU_obj.log_info("Invalid mode for Input Processor. " +\
                                      "Defaulting to mode 0(zero input)")
            self.record()
            return

        self.input_to_be_processed = True
        self.update_input()
        if self.memory_mode == 'cpu':
            for var in self.variables:
                self._d_input[var].set(self.variables[var]['input'])
        #TODO, record should be done also when is_input_available = False
        self.record()


    def inject_input(self, var):
        if var not in self.variables: return
        if not self.input_to_be_processed: return
        buff = self.memory_manager.get_buffer(var)
        # dest_mem = garray.GPUArray((1,buff.size),buff.dtype,
        #                            gpudata=int(buff.gpudata)+\
        #                            buff.current*buff.ld*\
        #                            buff.dtype.itemsize)
        dest_mem = int(buff.gpudata)+buff.current*buff.ld*buff.dtype.itemsize
        # self.add_inds(self._d_input[var], dest_mem, self.dest_inds[var])
        self.add_inds(var, self._d_input[var].gpudata,
                      dest_mem)

    def record(self):
        if self.input_file is not None:
            if self.record_count == 0:
                if not self.is_input_available() and self.mode == 0:
                    for var, d in self.variables.items():
                        var_folder = '{}/data'.format(var)
                        if isinstance(var, str):
                            self.input_file_handle[var_folder].resize((self.input_file_handle[var_folder].shape[0]+1,
                                                                        len(d['uids'])))
                            self.input_file_handle[var_folder][-1,:] = np.zeros((1,d['input'].size))
                        elif isinstance(var, tuple):
                            n = len(var)
                            for i, ind_var in enumerate(var):
                                var_subfolder = '{}/{}'.format(var_folder, ind_var)
                                self.input_file_handle[var_subfolder].resize(
                                                (self.input_file_handle[var_subfolder].shape[0]+1,
                                                                            len(d['uids'])))
                                self.input_file_handle[var_subfolder][-1,:] = np.zeros((1,d['input'][i::n].size))
                else:
                    for var, d in self.variables.items():
                        var_folder = '{}/data'.format(var)
                            # if var == 'spike_state':
                            #     self.input_file_handle[var_folder].resize((self.input_file_handle[var_folder].shape[0]+1,
                            #                                                 len(d['uids'])))
                            #     if self.memory_mode == 'cpu':
                            #         self.input_file_handle[var_folder][-1,:] = d['input'].reshape((1,-1))
                            #     elif self.memory_mode == 'gpu':
                            #         self.input_file_handle[var_folder][-1,:] = self._d_input[var].get().reshape((1,-1))
                            # else:
                        if isinstance(var, str):
                            self.input_file_handle[var_folder].resize((self.input_file_handle[var_folder].shape[0]+1,
                                                                        len(d['uids'])))
                            if self.memory_mode == 'cpu':
                                self.input_file_handle[var_folder][-1,:] = d['input'].reshape((1,-1))
                            elif self.memory_mode == 'gpu':
                                self.input_file_handle[var_folder][-1,:] = self._d_input[var].get().reshape((1,-1))
                        elif isinstance(var, tuple):
                            n = len(var)
                            for i, ind_var in enumerate(var):
                                var_subfolder = '{}/{}'.format(var_folder, ind_var)
                                self.input_file_handle[var_subfolder].resize(
                                                (self.input_file_handle[var_subfolder].shape[0]+1,
                                                                            len(d['uids'])))
                                # self.input_file_handle[var_subfolder][-1,:] = d['input'][i::n].reshape((1,-1))
                                if self.memory_mode == 'cpu':
                                    self.input_file_handle[var_subfolder][-1,:] = d['input'][i::n].reshape((1,-1))
                                elif self.memory_mode == 'gpu':
                                    self.input_file_handle[var_subfolder][-1,:] = self._d_input[var].get()[i::n].reshape((1,-1))
            self.record_count = (self.record_count + 1) % self.input_interval

    # Should be implemented by child class
    def update_input(self):
        raise NotImplementedError

    # Should be implemented by child class
    def is_input_available(self):
        raise NotImplementedError

    def _pre_run(self):
        assert(self.LPU_obj)
        assert all([var in self.memory_manager.variables
                    for var in self.variables.keys()]),\
               (list(self.memory_manager.variables), list(self.variables.keys()))
        self.add_inds_func = {}

        if self.input_file is not None:
            self.input_file_handle = h5py.File(self.input_file, 'w')
            self.input_file_handle.create_dataset('metadata',(),'i')
            self.input_file_handle['metadata'].attrs['dt'] = self.dt
            self.input_file_handle['metadata'].attrs['sample_interval'] = self.input_interval
            self.input_file_handle['metadata'].attrs['DateCreated'] = datetime.now().isoformat()
            for k, v in self.metadata:
                self.input_file_handle['metadata'].attrs[k] = v

        for var, d in self.variables.items():
            v_dict =  self.memory_manager.variables[var]
            uids = []
            inds = []
            for uid in d['uids']:
                cd = self.LPU_obj.conn_dict[uid]
                assert(var in cd)
                pre = cd[var]['pre'][0]
                inds.append(v_dict['uids'][pre])

            if isinstance(var, str):
                self.dest_inds[var] = garray.to_gpu(np.array(inds,np.int32))
                self.dtypes[var] = v_dict['buffer'].dtype
                self._d_input[var] = garray.zeros(len(d['uids']),self.dtypes[var])
                if self.memory_mode == 'cpu':
                    self.variables[var]['input'] = cuda.pagelocked_zeros(
                                                            len(d['uids']),
                                                            self.dtypes[var])
                elif self.memory_mode == 'gpu':
                    self.variables[var]['input'] = self._d_input[var]
                self.add_inds_func[var] = get_inds_kernel(self.dest_inds[var].dtype,
                                                          v_dict['buffer'].dtype)

            elif isinstance(var, tuple):
                n = len(var)
                new_inds = list(itertools.chain.from_iterable(
                                [[ind*n+i for i in range(n)] for ind in inds]))
                self.dest_inds[var] = garray.to_gpu(np.array(new_inds, np.int32))
                self.dtypes[var] = v_dict['buffer'].dtype
                self._d_input[var] = garray.zeros(len(d['uids'])*n, self.dtypes[var])
                if self.memory_mode == 'cpu':
                    self.variables[var]['input'] = cuda.pagelocked_zeros(
                                                            len(d['uids'])*n,
                                                            self.dtypes[var])
                elif self.memory_mode == 'gpu':
                    self.variables[var]['input'] = self._d_input[var]
                self.add_inds_func[var] = get_inds_kernel(self.dest_inds[var].dtype,
                                                          v_dict['buffer'].dtype)

            else:
                raise TypeError('variable name must either be a str or a tuple of str')


            if self.input_file is not None:
                self.input_file_handle.create_dataset('{}/uids'.format(var),
                                                      data=np.array(d['uids'],
                                                                    dtype = 'S'))
                if isinstance(var, str):
                    self.input_file_handle.create_dataset('{}/data'.format(var), (0,len(d['uids'])),
                                                          d['input'].dtype,
                                                          maxshape=(None,len(d['uids'])))
                elif isinstance(var, tuple):
                    n = len(var)
                    for ind_var in var:
                        self.input_file_handle.create_dataset('{}/data/{}'.format(var, ind_var), (0,len(d['uids'])),
                                                              d['input'].dtype,
                                                              maxshape=(None,len(d['uids'])))
                else:
                    raise TypeError('variable name must either be a str or a tuple of str')
        self.record_count = 0

        self.pre_run()

    def pre_run(self):
        pass

    def post_run(self):
        pass

    # def add_inds(self, src, dest, inds, dest_shift=0):
    #     """
    #     Set `dest[inds[i]+dest_shift] = src[i] for i in range(len(inds))`
    #     """
    #
    #     assert src.dtype == dest.dtype
    #     try:
    #         func = self.add_inds.cache[(inds.dtype, src.dtype)]
    #     except KeyError:
    #         inds_ctype = dtype_to_ctype(inds.dtype)
    #         data_ctype = dtype_to_ctype(src.dtype)
    #         v = ("{data_ctype} *dest, int dest_shift," +\
    #              "{inds_ctype} *inds, {data_ctype} *src").format(\
    #                     data_ctype=data_ctype,inds_ctype=inds_ctype)
    #         func = elementwise.ElementwiseKernel(v,\
    #         "dest[inds[i]+dest_shift] = dest[inds[i]+dest_shift] + src[i]")
    #         self.add_inds.cache[(inds.dtype, src.dtype)] = func
    #     func(dest, int(dest_shift), inds, src, range=slice(0, len(inds), 1) )

    # add_inds.cache = {}

    def add_inds(self, var, src, dest, dest_shift = 0):
        """
        Set `dest[inds[i]+dest_shift] = src[i] for i in range(len(inds))`
        """
        func = self.add_inds_func[var]
        inds = self.dest_inds[var]
        func.prepared_async_call(
            func.grid, func.block, None,
            dest, int(dest_shift), inds.gpudata, src, inds.size)

    def close_files(self):
        if self.input_file is not None:
            try:
                self.input_file_handle.close()
            except:
                pass

    def __del__(self):
        self.close_files()

    def set_metadata(self, metadata):
        """
        metadata: dictionary of metadata
        """
        self.metadata.update(metadata)

    @property
    def record_settings(self):
        return {'sensory_file': self.sensory_file,
                'sensory_interval': self.sensory_interval,
                'input_file': self.input_file,
                'input_interval': self.input_interval}


@context_dependent_memoize
def get_inds_kernel(inds_dtype, src_dtype):
    template = """
__global__ void update(%(data_ctype)s* dest, int dest_shift,
                       %(inds_ctype)s* inds, %(data_ctype)s* src, int N)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(inds_ctype)s ind;
    for(int i = tid; i < N; i += total_threads)
    {
        ind = inds[i]+dest_shift;
        dest[ind] += src[i];
    }
}
"""
    mod = SourceModule(template % {"data_ctype": dtype_to_ctype(src_dtype),
                                   "inds_ctype": dtype_to_ctype(inds_dtype)})
    func = mod.get_function("update")
    func.prepare('PiPPi')
    func.block = (128,1,1)
    func.grid = (16 * cuda.Context.get_device().MULTIPROCESSOR_COUNT, 1)
    return func



from neurokernel.LPU.NDComponents.AxonHillockModels.BaseAxonHillockModel import *

class ConnorStevens(BaseAxonHillockModel):
    updates = ['spike_state', # (bool)
               'V' # Membrane Potential (mV)
              ]
    accesses = ['I'] # (\mu A/cm^2 )
    params = ['n', # state variable for activation of K channel ([0-1] unitless)
              'm', # state variable for activation of Na channel ([0-1] unitless)
              'h', # state variable for inactivation of Na channel ([0-1] unitless)
              'a', # state variable for activation of A channel ([0-1] unitless)
              'b' # state variable for inactivation of A channel ([0-1] unitless)
              ]
    extra_params = ['initV']
    internals = OrderedDict([('V',-65.), # Membrane Potential (mV)
                             ('Vprev1',-65.) # Membrane Potential (mV)
                            ])

    @property
    def maximum_dt_allowed(self):
        return 1e-5

    def pre_run(self, update_pointers):
        super(ConnorStevens, self).pre_run(update_pointers)
        # if 'initV' in self.params_dict:
        self.add_initializer('initV', 'Vprev1', update_pointers)

    def get_update_template(self):
        template = """
#define EXP exp%(fletter)s
#define POW pow%(fletter)s

#define		E_K		-72.
#define		E_Na		55.
#define		E_a		-75.
#define		E_l		-17.
#define		G_total		67.7
#define		G_a		47.7
#define		G_Na		120.
#define		G_K		(G_total-G_a)
#define		G_l		0.3
#define		ms		-5.3
#define		hs		-12.
#define		ns		-4.3

__global__ void update(
    int num_comps,
    %(dt)s dt,
    int nsteps,
    %(input_I)s* g_I,
    %(param_n)s* g_n,
    %(param_m)s* g_m,
    %(param_h)s* g_h,
    %(param_a)s* g_a,
    %(param_b)s* g_b,
    %(internal_V)s* g_Vprev,
    %(internal_Vprev1)s* g_Vprev1,
    %(update_spike_state)s* g_spike_state,
    %(update_V)s* g_V)
{
	int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(dt)s ddt = dt*1000.; // s to ms

    %(update_V)s V, Vprev1, Vprev2;
    %(input_I)s I;
    %(update_spike_state)s spike;

    %(param_n)s n, a_n, b_n, n_inf, tau_n;
    %(param_m)s m, a_m, b_m, m_inf, tau_m;
    %(param_h)s h, a_h, b_h, h_inf, tau_h;
    %(param_a)s a, a_inf, tau_a;
    %(param_b)s b, b_inf, tau_b;


    for(int i = tid; i < num_comps; i += total_threads)
    {
        spike = 0;
        I = g_I[i];
        n = g_n[i];
        m = g_m[i];
        h = g_h[i];
        a = g_a[i];
        b = g_b[i];
        V = g_Vprev[i];
        Vprev1 = V;
        Vprev2 = g_Vprev1[i];

        for (int j = 0; j < nsteps; ++j)
        {
            /*
             * Hodgkin-Huxley with shifts - 3.8 is temperature factor
             */
            a_m = -.1*(V+35+ms)/(EXP(-(V+35+ms)/10)-1);
            b_m = 4*EXP(-(V+60+ms)/18);
            m_inf = a_m/(a_m+b_m);
            tau_m = 1/(3.8*(a_m+b_m));

            a_h = .07*EXP(-(V+60+hs)/20);
            b_h = 1/(1+EXP(-(V+30+hs)/10));
            h_inf = a_h/(a_h+b_h);
            tau_h = 1/(3.8*(a_h+b_h));

            a_n = -.01*(V+50+ns)/(EXP(-(V+50+ns)/10)-1);
            b_n = .125*EXP(-(V+60+ns)/80);
            n_inf = a_n/(a_n+b_n);
            tau_n = 2/(3.8*(a_n+b_n));

            a_inf = POW(.0761*EXP((V+94.22)/31.84)/(1+EXP((V+1.17)/28.93)),.3333);
            tau_a = .3632+1.158/(1+EXP((V+55.96)/20.12));
            b_inf = POW(1/(1+EXP((V+53.3)/14.54)),4);
            tau_b = 1.24+2.678/(1+EXP((V+50)/16.027));

            V += ddt*(I-G_l*(V-E_l)-G_Na*h*m*m*m*(V-E_Na)-G_K*n*n*n*n*(V-E_K)-G_a*b*a*a*a*(V-E_a));
            m += ddt*(m_inf-m)/tau_m;
            h += ddt*(h_inf-h)/tau_h;
            n += ddt*(n_inf-n)/tau_n;
            a += ddt*(a_inf-a)/tau_a;
            b += ddt*(b_inf-b)/tau_b;

            spike += (Vprev2<=Vprev1) && (Vprev1 >= V) && (Vprev1 > -30);

            Vprev2 = Vprev1;
            Vprev1 = V;

        }
        g_n[i] = n;
        g_m[i] = m;
        g_h[i] = h;
        g_a[i] = a;
        g_b[i] = b;
        g_V[i] = V;
        g_Vprev[i] = V;
        g_Vprev1[i] = Vprev2;
        g_spike_state[i] = (spike > 0);
    }
}
"""
        return template

if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.FileInputProcessor import FileInputProcessor
    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 0.3
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('neuron0', **{
               'class': 'ConnorStevens',
               'name': 'ConnorStevens',
               'n': 0.,
               'm': 0.,
               'h': 1.,
               'a': 0.,
               'b': 0.,
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    fl_input_processor = StepInputProcessor('I', ['neuron0'], 40, 0.15, 0.25)
    fl_output_processor = FileOutputProcessor([('spike_state', None),('V', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()

    # plot the result
    import h5py
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt

    f = h5py.File('new_output.h5')
    t = np.arange(0, args.steps)*dt

    plt.figure()
    plt.subplot(211)
    plt.plot(t,list(f['V'].values())[0])
    plt.xlabel('time, [s]')
    plt.ylabel('Voltage, [mV]')
    plt.title('Connor-Stevens Neuron')
    plt.xlim([0, dur])
    plt.ylim([-70, 60])
    plt.grid()
    plt.subplot(212)
    spk = f['spike_state/data'].value.flatten().nonzero()[0]
    plt.stem(t[spk],np.ones((len(spk),)))
    plt.xlabel('time, [s]')
    plt.ylabel('Spike')
    plt.xlim([0, dur])
    plt.ylim([0, 1.2])
    plt.grid()
    plt.tight_layout()
    plt.savefig('csn.png',dpi=300)



#!/usr/bin/env python
import warnings
from abc import ABCMeta, abstractmethod, abstractproperty
import os.path
import numpy as np

import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from neurokernel.LPU.utils.simpleio import *
from .BaseDendriteModel import BaseDendriteModel

class Aggregator(BaseDendriteModel):
    accesses = ['g','V']
    updates = ['I']

    def __init__(self, params_dict, access_buffers, dt, debug=False,
                 LPU_id=None, cuda_verbose=False):
        if cuda_verbose:
            self.compile_options = ['--ptxas-options=-v']
        else:
            self.compile_options = []

        self.LPU_id = LPU_id
        self.access_buffers = access_buffers
        self.params_dict = params_dict
        self.debug = debug
        self.dt = dt
        self.LPU_id = LPU_id

        self.num_comps = params_dict['pre']['V'].size


        self.update = self.get_update_func(self.access_buffers['g'].dtype)

    @property
    def maximum_dt_allowed(self):
        return self.dt

    def run_step(self, update_pointers, st=None):
        self.update.prepared_async_call(\
                        self.grid, self.block, st,
                        self.access_buffers['g'].gpudata,                     #P
                        self.access_buffers['g'].ld,                          #i
                        self.access_buffers['g'].current,                     #i
                        self.access_buffers['g'].buffer_length,               #i
                        self.params_dict['conn_data']['g']['delay'].gpudata,  #P
                        self.params_dict['conn_data']['g']['reverse'].gpudata,#P
                        self.params_dict['pre']['g'].gpudata,                 #P
                        self.params_dict['npre']['g'].gpudata,                #P
                        self.params_dict['cumpre']['g'].gpudata,              #P
                        self.access_buffers['V'].gpudata,                     #P
                        self.access_buffers['V'].ld,                          #i
                        self.access_buffers['V'].current,                     #i
                        self.params_dict['pre']['V'].gpudata,                 #P
                        update_pointers['I'])                                 #P


    def get_update_func(self, dtype=np.double):
        template = """
        #define NUM_COMPS %(num_comps)d

        __global__ void aggregate_I(%(type)s* g, int ld, int current,
                                    int buffer_length, int* delay,
                                    %(type)s* V_rev, int* pre, int* npre,
                                    int* cumpre, %(type)s* V, int V_ld,
                                    int V_current, int* V_pre, %(type)s* I)
        {
            // must use block size (32, 32, 1)
            int tidx = threadIdx.x;
            int tidy = threadIdx.y;
            int bid = blockIdx.x;
            int comp;

            __shared__ int num_pre[32];
            __shared__ int pre_start[32];
            __shared__ double V_in[32];
            __shared__ double input[32][33];

            if(tidy == 0)
            {
                comp = bid * 32 + tidx;
                if(comp < NUM_COMPS)
                {
                    num_pre[tidx] = npre[comp];
                    V_in[tidx] = V[V_pre[comp]+V_current*V_ld];
                }
            } else if(tidy == 1)
            {
                comp = bid * 32 + tidx;
                if(comp < NUM_COMPS)
                {
                    pre_start[tidx] = cumpre[comp];
                    I[comp] = 0;
                }
            }

            input[tidy][tidx] = 0.0;

            __syncthreads();

            comp = bid * 32 + tidy ;
            if(comp < NUM_COMPS)
            {
               int dl;
               int col;
               int n_pre = num_pre[tidy];
               int start = pre_start[tidy];
               double VV = V_in[tidy];


               for(int i = tidx; i < n_pre; i += 32)
               {
                   dl = delay[start+i];
                   col = current - dl;
                   if(col < 0)
                   {
                     col = buffer_length + col;
                   }

                   input[tidy][tidx] += g[pre[start + i] + col*ld] *
                                        (VV - V_rev[start + i]);
               }
            }

            __syncthreads();

            if(tidy < 8)
            {
                input[tidx][tidy] += input[tidx][tidy + 8];
                input[tidx][tidy] += input[tidx][tidy + 16];
                input[tidx][tidy] += input[tidx][tidy + 24];
            }

            __syncthreads();

            if(tidy < 4)
            {
                input[tidx][tidy] += input[tidx][tidy + 4];
            }

            __syncthreads();

            if(tidy < 2)
            {
                input[tidx][tidy] += input[tidx][tidy + 2];
            }

            __syncthreads();

            if(tidy == 0)
            {
                input[tidx][0] += input[tidx][1];
                comp = bid*32+tidx;
                if(comp < NUM_COMPS)
                {
                    I[comp] -= input[tidx][0];
                }
            }
        }
        """
        mod = SourceModule(template % {"num_comps": self.num_comps,
                                       "type": dtype_to_ctype(dtype)},
                           options=self.compile_options)
        func = mod.get_function("aggregate_I")
        func.prepare('PiiiPPPPPPiiPP')
        self.block = (32, 32, 1)
        self.grid = ((self.num_comps - 1) // 32 + 1, 1)
        return func



#!/usr/bin/env python

"""
LPU output visualization.
"""

import collections
from collections import OrderedDict
import itertools
import os
import h5py

from future.utils import itervalues, iteritems
import matplotlib
from matplotlib import cm
from matplotlib.colors import Normalize
import matplotlib.pyplot as plt
plt.ioff() # interactive mode can interfere with frame updates
from matplotlib.animation import FFMpegFileWriter, AVConvFileWriter
from matplotlib.colors import hsv_to_rgb
import networkx as nx
import numpy as np
from scipy.interpolate import griddata
from shutilwhich import which

import neurokernel.LPU.utils.simpleio as sio

class visualizer(object):
    """
    Visualize the output produced by LPU models.

    Examples
    --------
    >>> import neurokernel.LPU.utils.visualizer as vis
    >>> V = vis.visualizer()
    >>> config1 = {}
    >>> config1['type'] = 'image'
    >>> config1['shape'] = [32,24]
    >>> config1['clim'] = [-0.6,0.5]
    >>> config2 = config1.copy()
    >>> config2['clim'] = [-0.55,-0.45]
    >>> V.add_LPU('lamina_output.h5', 'lamina.gexf.gz','lamina')
    >>> V.add_plot(config1, 'lamina', 'R1')
    >>> V.add_plot(config2, 'lamina', 'L1')
    >>> V.update_interval = 50
    >>> V.out_filename = 'test.avi'
    >>> V.run()
    """

    def __init__(self):
        self._xlim = [0,1]
        self._ylim = [-1,1]
        self._imlim = [-1, 1]
        self._update_interval = -1
        self._out_file = None
        self._fps = 5
        self._codec = 'libtheora'
        self._config = OrderedDict()
        self._rows = 0
        self._cols = 0
        self._figsize = (16,9)
        self._fontsize = 18
        self._t = 0
        self._dts = {}
        self._sample_intervals = {}
        self._start_times = {}
        self._data = {}
        self._uids = {}
        self._maxt = None
        self._title = None
        self._FFMpeg = None

    def add_LPU(self, data_file, LPU='', win=None, is_input=False,gexf_file=None,
                sample_interval=1, start_time=0, dt=1e-4, transpose_axes = [1,0]):
        """
        Add data associated with a specific LPU to a visualization.

        To add a plot containing neurons from a particular LPU,
        the LPU needs to be added to the visualization using this
        function. Note that outputs from multiple neurons can
        be visualized using the same visualizer object. The IDs
        specified in the arguments passed to `add_plot()` are assumed to be
        indices into array stored in the HDF5 file.

        Parameters
        ----------
        data_file : str
             Location of the HDF5 file generated by neurokernel
             containing the output of the LPU
        LPU : str
            Name of the LPU. Will be used as identifier to add plots.
            For input signals, the name of the LPU will be prepended
            with 'input_'. For example::

                V.add_LPU('vision_in.h5', LPU='vision')

            will create the LPU identifier 'input_vision'.
            Therefore, adding a plot depicting this input can be done by::

                V.add_plot({''type':'image',imlim':[-0.5,0.5]},LPU='input_vision)
        win : slice/list (Optional)
            Can be used to limit the visualization to a specific time window.
        gexf_file : string (Optional)
            Location of gexf file containg the graph of the LPU configuration
        graph : networkx graph object(Otional)
            NetworkX graph object representing the LPU configuration
        is_input : Boolean (optional)
            Set to true if the data_file represents input
        dt, start_time, sample_interval: double, double, int (All Optional)
            These arguments will only be used to set these attributes for
            input h5 files. For all other cases, these will be read from
            the h5 file

        All arguments beyond LPU should be considered strictly keyword only
        """

        if is_input:
            LPU = 'input_' + str(LPU)
            self._sample_intervals[LPU] = sample_interval
            self._dts[LPU] = dt * sample_interval
            self._start_times[LPU] = start_time
            f = h5py.File(data_file)
            self._uids[LPU] = {}
            self._data[LPU] = {}
            for k, d in iteritems(f):
                self._uids[LPU][k] = [a.decode('utf8') for a in f[k]['uids'].value]
                self._data[LPU][k] = np.transpose(f[k]['data'].value,
                                                  axes = transpose_axes)

            self._config[LPU] = []
            if self._maxt:
                self._maxt = min(self._maxt,
                                 (self._data[LPU][k].shape[1]-1)*self._dts[LPU])
            else:
                self._maxt = (self._data[LPU][k].shape[1]-1)*self._dts[LPU]
            f.close()
            return

        self._config[LPU] = []
        f = h5py.File(data_file)

        self._sample_intervals[LPU] = f['metadata'].attrs['sample_interval']
        self._dts[LPU] = f['metadata'].attrs['dt'] * self._sample_intervals[LPU]
        self._start_times[LPU] = f['metadata'].attrs['start_time']
        self._uids[LPU] = {}
        self._data[LPU] = {}
        for k, d in iteritems(f):
            if k=='metadata': continue
            self._uids[LPU][k] = [a.decode('utf8') for a in f[k]['uids'].value]
            self._data[LPU][k] = np.transpose(f[k]['data'].value,
                                              axes = transpose_axes)

        if win is not None:
            for k in self._data[LPU]:
                self._data[LPU][k] = self._data[LPU][k][:,win]
        k = list(self._data[LPU])[0]
        if self._maxt:
            self._maxt = min(self._maxt,
                             (self._data[LPU][k].shape[1]-1)*self._dts[LPU])
        else:
            self._maxt = (self._data[LPU][k].shape[1]-1)*self._dts[LPU]
        f.close()

    def run(self, final_frame_name=None, dpi=300):
        """
        Starts the visualization process.

        If the property `out_filename` is set, the visualization is saved as a
        video to the disk; if not, the animation is displayed on screen.
        Please refer to documentation of `add_LPU`, `add_plot`
        and the properties of this class on how to configure the visualizer
        before calling this method. An example can be found in the class doc
        string.

        Parameters
        ----------
        final_frame_name : str, optional
            If specified, the final frame of the animation is saved
            to disk.
        dpi : int, default=300
            Resolution at which final frame is saved to disk if
            `final_frame_name` is specified.

        Notes
        -----
        If `update_interval` is set to 0 or None, it will be replaced by the
        index of the final time step. As a result, the visualizer will only
        generate and save the final frame if `final_frame_name` is set.
        """

        self.final_frame_name = final_frame_name
        self._initialize()
        if not self._update_interval:
            self._update_interval = self._maxt
            self._t = self._maxt
        if self._update_interval == -1:
            self._update_interval = max(np.asarray(list(self._dts.values())))*50
        for _ in np.arange(self._t,self._maxt*(1+np.finfo(float).eps),
                           self._update_interval):
            self._update()
        if final_frame_name is not None:
            self.f.savefig(final_frame_name, dpi=dpi)
        if self.out_filename:
            self._close()

    def _set_wrapper(self, obj, name, value):
        name = name.lower()
        func = getattr(obj, 'set_'+name, None)
        if func:
            try:
                func(value, fontsize=self._fontsize, weight='bold')
            except:
                try:
                    func(value)
                except:
                    pass

    def _initialize(self):

        # Count number of plots to create:
        num_plots = 0
        for config in itervalues(self._config):
            num_plots += len(config)

        # Set default grid of plot positions:
        if not self._rows*self._cols == num_plots:
            self._cols = int(np.ceil(np.sqrt(num_plots)))
            self._rows = int(np.ceil(num_plots/float(self._cols)))
        self.f, self.axarr = plt.subplots(self._rows, self._cols,
                                          figsize=self._figsize)

        # Remove unused subplots:
        for i in range(num_plots, self._rows*self._cols):
            plt.delaxes(self.axarr[np.unravel_index(i, (self._rows, self._cols))])
        cnt = 0
        self.handles = []
        self.types = []
        keywds = ['handle', 'ydata', 'fmt', 'type', 'ids', 'shape', 'norm']
        # TODO: Irregular grid in U will make the plot better
        U, V = np.mgrid[0:np.pi/2:complex(0, 60),
                        0:2*np.pi:complex(0, 60)]
        X = np.cos(V)*np.sin(U)
        Y = np.sin(V)*np.sin(U)
        Z = np.cos(U)
        self._dome_pos_flat = (X.flatten(), Y.flatten(), Z.flatten())
        self._dome_pos = (X, Y, Z)
        self._dome_arr_shape = X.shape
        if not isinstance(self.axarr, np.ndarray):
            self.axarr = np.asarray([self.axarr])
        for LPU, configs in iteritems(self._config):
            dt = self._dts[LPU]
            for plt_id, config in enumerate(configs):
                var = config['variable']
                ind = np.unravel_index(cnt, self.axarr.shape)
                cnt+=1
                # Some plot types require specific numbers of
                # neuron ID arrays:
                if 'type' in config:
                    if config['type'] == 'quiver':
                        assert len(config['ids'])==2
                        config['type'] = 0
                    elif config['type'] == 'hsv':
                        assert len(config['ids'])==2
                        config['type'] = 1
                    elif config['type'] == 'image':
                        assert len(config['ids'])==1
                        config['type'] = 2
                    elif config['type'] == 'waveform':
                        config['type'] = 3
                    elif config['type'] == 'raster':
                        config['type'] = 4
                    elif config['type'] == 'rate':
                        config['type'] = 5
                    elif config['type'] == 'dome':
                        config['type'] = 6
                        assert('lat' in config and 'long' in config)
                    else:
                        raise ValueError('Plot type not supported')
                else:
                    if (str(LPU).startswith('input') and not
                        self._graph[LPU].node[str(config['ids'][0][0])]['spiking']):
                        config['type'] = 2
                    else:
                        config['type'] = 3

                if config['type'] < 3:
                    if not 'shape' in config:
                        num_neurons = len(config['ids'][0])
                        config['shape'] = [int(np.ceil(np.sqrt(num_neurons)))]
                        config['shape'].append(int(np.ceil(num_neurons/float(config['shape'][0]))))

                if config['type'] == 0:
                    config['handle'] = self.axarr[ind].quiver(\
                               np.reshape(self._data[LPU][var][config['ids'][0],0],config['shape']),\
                               np.reshape(self._data[LPU][var][config['ids'][1],0],config['shape']))
                elif config['type'] == 1:
                    X = np.reshape(self._data[LPU][var][config['ids'][0],0],config['shape'])
                    Y = np.reshape(self._data[LPU][var][config['ids'][1],0],config['shape'])
                    V = (X**2 + Y**2)**0.5
                    H = (np.arctan2(X,Y)+np.pi)/(2*np.pi)
                    S = np.ones_like(V)
                    HSV = np.dstack((H,S,V))
                    RGB = hsv_to_rgb(HSV)
                    config['handle'] = self.axarr[ind].imshow(RGB)
                elif config['type'] == 2:
                    if 'trans' in config:
                        if config['trans'] is True:
                            to_transpose = True
                        else:
                            to_transpose = False
                    else:
                        to_transpose = False
                        config['trans'] = False

                    if to_transpose:
                        temp = self.axarr[ind].imshow(np.transpose(np.reshape(\
                                self._data[LPU][var][config['ids'][0],0], config['shape'])))
                    else:
                        temp = self.axarr[ind].imshow(np.reshape(\
                                self._data[LPU][var][config['ids'][0],0], config['shape']))

                    temp.set_clim(self._imlim)
                    temp.set_cmap(plt.cm.gist_gray)
                    config['handle'] = temp
                elif config['type'] == 3:
                    fmt = config['fmt'] if 'fmt' in config else ''
                    self.axarr[ind].set_xlim(self._xlim)
                    self.axarr[ind].set_ylim(self._ylim)
                    if len(config['ids'][0])==1:
                        config['handle'] = self.axarr[ind].plot([0], \
                                            [self._data[LPU][var][config['ids'][0][0],0]], fmt)[0]
                        config['ydata'] = [self._data[LPU][var][config['ids'][0][0],0]]
                    else:
                        config['handle'] = self.axarr[ind].plot(self._data[LPU][var][config['ids'][0],0])[0]

                elif config['type'] == 4:
                    config['handle'] = self.axarr[ind]
                    config['handle'].vlines(0, 0, 0.01)
                    config['handle'].set_ylim([.5, len(config['ids'][0]) + .5])
                    config['handle'].set_ylabel('Neurons',
                                                fontsize=self._fontsize-1, weight='bold')
                    config['handle'].set_xlabel('Time (s)',fontsize=self._fontsize-1, weight='bold')

                    config['handle'].set_xlim([0,self._data[LPU][var].shape[1]*dt])
                    config['handle'].axes.set_yticks([])
                    eps = np.finfo(float).eps
                    config['handle'].axes.set_xticks(\
                            np.linspace(0+self._start_times[LPU],
                                        self._start_times[LPU]+self._maxt,11))
                elif config['type'] == 6:
                    self.axarr[ind].axes.set_yticks([])
                    self.axarr[ind].axes.set_xticks([])
                    self.axarr[ind] = self.f.add_subplot(self._rows,
                                                         self._cols,
                                                         cnt,
                                                         projection='3d')
                    config['handle' ] = self.axarr[ind]
                    config['handle'].axes.set_yticks([])
                    config['handle'].axes.set_xticks([])
                    config['handle'].xaxis.set_ticks([])
                    config['handle'].yaxis.set_ticks([])
                    config['handle'].zaxis.set_ticks([])
                    if 'norm' not in config:
                        config['norm'] = Normalize(vmin=-70, vmax=0, clip=True)
                    elif config['norm'] == 'auto':
                        if self._data[LPU][var].shape[1] > 100:
                            config['norm'] = Normalize(vmin = np.min(self._data[LPU][var][config['ids'][0],100:]),
                                                       vmax = np.max(self._data[LPU][var][config['ids'][0],100:]),
                                                       clip = True)
                        else:
                            config['norm'] = Normalize(vmin = np.min(self._data[LPU][var][config['ids'][0],:]),
                                                       vmax = np.max(self._data[LPU][var][config['ids'][0],:]),
                                                       clip = True)

                    node_dict = self._graph[LPU].node
                    latpositions = config['lat']
                    longpositions = config['long']
                    xx = np.cos(longpositions) * np.sin(latpositions)
                    yy = np.sin(longpositions) * np.sin(latpositions)
                    zz = np.cos(latpositions)
                    config['positions'] = (xx, yy, zz)
                    colors = griddata(config['positions'], self._data[LPU][var][config['ids'][0],0],
                                      self._dome_pos_flat, 'nearest').reshape(self._dome_arr_shape)
                    colors = config['norm'](colors).data
                    colors = np.tile(np.reshape(colors,
                                                [self._dome_arr_shape[0],self._dome_arr_shape[1],1])
                                     ,[1,1,4])
                    colors[:,:,3] = 1.0
                    config['handle'].plot_surface(self._dome_pos[0], self._dome_pos[1],
                                                  self._dome_pos[2], rstride=1, cstride=1,
                                                  facecolors=colors, antialiased=False,
                                                  shade=False)

                for key in config:
                    if key not in keywds:
                        try:
                            self._set_wrapper(self.axarr[ind],key, config[key])
                        except:
                            pass
                        try:
                            self._set_wrapper(config['handle'],key, config[key])
                        except:
                            pass

                if config['type']<3:
                    config['handle'].axes.set_xticks([])
                    config['handle'].axes.set_yticks([])

            if self.suptitle is not None:
                self.f.suptitle(self._title, fontsize=self._fontsize+1, x=0.5,y=0.03, weight='bold')

        plt.tight_layout()

        if self.out_filename and self.update_interval:
            if self.FFMpeg is None:
                if which(matplotlib.rcParams['animation.ffmpeg_path']):
                    self.writer = FFMpegFileWriter(fps=self.fps, codec=self.codec)
                elif which(matplotlib.rcParams['animation.avconv_path']):
                    self.writer = AVConvFileWriter(fps=self.fps, codec=self.codec)
                else:
                    raise RuntimeError('cannot find ffmpeg or avconv')
            elif self.FFMpeg:
                if which(matplotlib.rcParams['animation.ffmpeg_path']):
                    self.writer = FFMpegFileWriter(fps=self.fps, codec=self.codec)
                else:
                    raise RuntimeError('cannot find ffmpeg')
            else:
                if which(matplotlib.rcParams['animation.avconv_path']):
                    self.writer = AVConvFileWriter(fps=self.fps, codec=self.codec)
                else:
                    raise RuntimeError('cannot find avconv')

            # Use the output file to determine the name of the temporary frame
            # files so that two concurrently run visualizations don't clobber
            # each other's frames:
            self.writer.setup(self.f, self.out_filename, dpi=80,
                              frame_prefix=os.path.splitext(self.out_filename)[0]+'_')
            self.writer.frame_format = 'png'
            self.writer.grab_frame()
        elif not self.final_frame_name:
            self.f.show()

    def _update(self):
        t = self._t
        for LPU, configs in iteritems(self._config):
            dt = self._dts[LPU]
            for config in configs:
                var = config['variable']
                data = self._data[LPU][var]
                if config['type'] == 3:
                    if round(float(t)/dt) >= data.shape[1] : continue
                    if len(config['ids'][0])==1:
                        s = int(round(float(t-self._update_interval)/dt))
                        e = int(round(float(t)/dt))
                        config['ydata'].extend(
                            np.reshape(np.double(data[config['ids'][0], \
                                        max(0,s):min(e,data.shape[1])]),(-1,)))
                        config['handle'].set_xdata(dt* np.arange(0, \
                                len(config['ydata']))+self._start_times[LPU])
                        config['handle'].set_ydata(np.asarray(config['ydata']))
                    else:
                        config['handle'].set_ydata(data[config['ids'][0],
                                                        int(round(float(t)/dt))])
                elif config['type']==4:
                    if int(round(float(t)/dt)) >= data.shape[1] : continue
                    for j, id in enumerate((config['ids'][0])):
                        s = int(round(float(t-self._update_interval)/dt))
                        e = int(round(float(t)/dt))
                        tmp = np.where(data[id, \
                                      max(0,s):min(e,data.shape[1])])[0]
                        for tind in tmp:
                            shift = self._start_times[LPU]
                            config['handle'].vlines(float(shift+(max(0,s)+tind)*dt),
                                                    j+0.75, j+1.25)
                elif config['type'] == 0:
                    if int(round(float(t)/dt)) >= data.shape[1] : continue
                    ind = int(round(float(t)/dt))
                    shape = config['shape']
                    ids = config['ids']
                    config['handle'].U = np.reshape(data[ids[0], ind],shape)
                    config['handle'].V = np.reshape(data[ids[1], ind],shape)
                elif config['type']==1:
                    if int(round(float(t)/dt)) >= data.shape[1] : continue
                    ind = int(round(float(t)/dt))
                    shape = config['shape']
                    ids = config['ids']
                    X = np.reshape(data[ids[0], ind],shape)
                    Y = np.reshape(data[ids[1], ind],shape)
                    V = (X**2 + Y**2)**0.5
                    H = (np.arctan2(X,Y)+np.pi)/(2*np.pi)
                    S = np.ones_like(V)
                    HSV = np.dstack((H,S,V))
                    RGB = hsv_to_rgb(HSV)
                    config['handle'].set_data(RGB)
                elif config['type'] == 2:
                    if int(round(float(t)/dt)) >= data.shape[1] : continue
                    ind = int(round(float(t)/dt))
                    ids = config['ids']
                    if config['trans']:
                        config['handle'].set_data(
                            np.transpose(np.reshape(data[ids[0], ind],
                                                    config['shape'])))
                    else:
                        config['handle'].set_data(
                            np.reshape(data[ids[0], ind], config['shape']))
                elif config['type'] == 6:
                    if int(round(float(t)/dt)) >= data.shape[1] : continue
                    ind = int(round(float(t)/dt))
                    ids = config['ids']
                    d = data[ids[0], ind]
                    colors = griddata(config['positions'], d,
                                      self._dome_pos_flat, 'nearest').reshape(self._dome_arr_shape)
                    colors = config['norm'](colors).data
                    colors = np.tile(np.reshape(colors,
                                                [self._dome_arr_shape[0],self._dome_arr_shape[1],1])
                                     ,[1,1,4])
                    colors[:,:,3] = 1.0

                    config['handle'].clear()
                    config['handle'].xaxis.set_ticks([])
                    config['handle'].yaxis.set_ticks([])
                    config['handle'].zaxis.set_ticks([])

                    config['handle'].plot_surface(self._dome_pos[0], self._dome_pos[1],
                                                  self._dome_pos[2], rstride=1, cstride=1,
                                                  facecolors=colors, antialiased=False,
                                                  shade=False)
                keywds = ['handle', 'ydata', 'fmt', 'type', 'ids', 'shape', 'norm']
                for key in config:
                    if key not in keywds:
                        try:
                            self._set_wrapper(self.axarr[ind],key, config[key])
                        except:
                            pass

                        try:
                            self._set_wrapper(config['handle'],key, config[key])
                        except:
                            pass
        self.f.canvas.draw()
        if self.out_filename:
            self.writer.grab_frame()

        self._t+=self._update_interval

    def add_plot(self, config_dict, LPU):
        """
        Add a plot to the visualizer

        Parameters
        ----------
        config_dict: dict
            A dictionary specifying the plot attributes. The attribute
            names should be the keys.

            The following are the plot attributes that can be specfied using
            this dict.

            type - str  (Required)
                This specifies the type of the plot. Has to be one of
                ['waveform', 'raster', 'image','hsv','quiver', 'dome']
                'dome' requires lat and long to specified. The size of lat and long
                should be the same as uids
            variable - str (Required)
                The variable to plot
            uids - list/tuple of either 1 or 2 list/tuple of uids
                Specifies the neuron uids from the associated LPU.
                The list of uids consistent with uids in the h5file.
                For example::

                    {'variable':'V','uids':[['1','2']]}

                will plot neurons with uids 1 and 2.
                Two lists of uids are needed if the plot is
                of type 'hsv' or 'quiver'
                For example::

                     {'variable':'V:,
                     'uids':[map(str,range(768)),map(str,range(768,1536)],'type':'HSV'}

                can be used to generate a HSV plot where the hue channel is
                controlled by the angle of the vector with the membrane
                potentials of the component with uids '0' through '767' as the x component
                and potentials of neurons with 'uids' '768' through '1535' as the y component.
                The value will be the magnitude of the same vector.

                This parameter must be specified for plots of type 'HSV' or 'quiver'

                If not specified, all components in the h5 file for which that variable was recorded
                will be plotted

            shape - list or tuple with two entries
                This attribute specifies the dimensions for plots of type image,
                hsv or quiver.

            title - str
                Optional. Can be used to control the title of the plot.


            In addition to the above, any parameter supported by matlpotlib
            for the particular type of plot can be specified.
            For example - 'imlim','clim','xlim','ylim' etc.
        LPU: str
            The name of the LPU associated to this plot.
        """

        config = config_dict.copy()
        assert('variable' in config)
        var = config['variable']
        assert(LPU in self._uids and LPU in self._data)
        if 'uids' in config:
            if not (isinstance(config['uids'], list) or
                    isinstance(config['uids'], np.ndarray)):
                config['uids'] = [config['uids']]
            config['ids'] = []
            for uids in config['uids']:
                tmp_list = []
                for uid in uids:
                    tmp = np.where(np.array(self._uids[LPU][var])==uid)[0]
                    if len(tmp):
                        tmp_list.append(tmp[0])
                config['ids'].append(tmp_list)
            self._config[LPU].append(config)
        elif str(LPU).startswith('input'):
            config['ids'] = [range(0, self._data[LPU][var].shape[0])]
            self._config[LPU].append(config)
        else:
            config['uids'] = self._uids[LPU][var]
            config['ids'] = [[range(0,len(config['uids']))]]
            self._config[LPU].append(config)
            #raise ValueError('uids must be provided')
            '''
            config['ids'] = {}
            for i,name in enumerate(names):
                config['ids'][i]=[]
                for id in range(len(self._graph[LPU].node)):
                    if self._graph[LPU].node[str(id)]['name'] == name:
                        config['ids'][i].append(id-shift)

            self._config[LPU].append(config)
            '''
        if not 'title' in config:
            if str(LPU).startswith('input_'):
                config['title'] = LPU.split('_',1)[1] + ' - ' + 'Input'
            else:
                config['title'] = str(LPU)

    def _close(self):
        self.writer.finish()
        plt.close(self.f)

    @property
    def xlim(self):
        """
        X-axis limits for all the raster and waveform plots. Can be superseded
        for individual plots by specifying xlim in the config_dict for that
        plot.

        See Also
        --------
        add_plot

        """
        return self._xlim

    @xlim.setter
    def xlim(self, value):
        self._xlim = value

    @property
    def ylim(self):
        """
        Get or set the limits of the y-axis for all the raster and waveform plots.
        Can be superseded for individual plots by specifying xlim in the config_dict
        for that plot.

        See Also
        --------
        add_plot
        """
        return self._ylim

    @ylim.setter
    def ylim(self, value):
        self._ylim = value

    @property
    def FFMpeg(self): return self._FFMpeg

    @FFMpeg.setter
    def FFMpeg(self, value):
        self._FFMpeg = value

    @property
    def imlim(self): return self._imlim

    @imlim.setter
    def imlim(self, value):
        self._imlim = value

    @property
    def out_filename(self): return self._out_file

    @out_filename.setter
    def out_filename(self, value):
        assert(isinstance(value, str))
        self._out_file = value

    @property
    def fps(self): return self._fps

    @fps.setter
    def fps(self, value):
        assert(isinstance(value, int))
        self._fps = value

    @property
    def codec(self): return self._codec

    @codec.setter
    def codec(self, value):
        assert(isinstance(value, str))
        self._codec = value

    @property
    def rows(self): return self._rows

    @rows.setter
    def rows(self, value):
        self._rows = value

    @property
    def cols(self): return self._cols

    @cols.setter
    def cols(self, value):
        self._cols = value

    @property
    def figsize(self): return self._figsize

    @figsize.setter
    def figsize(self, value):
        assert(isinstance(value, tuple) and len(value)==2)
        self._figsize = value

    @property
    def fontsize(self): return self._fontsize

    @fontsize.setter
    def fontsize(self, value):
        self._fontsize = value

    @property
    def suptitle(self): return self._title

    @suptitle.setter
    def suptitle(self, value):
        self._title = value

    @property
    def update_interval(self):
        """
        Update interval (in unit of time same as dt in the h5 files) for the animation.
        """
        return self._update_interval

    @update_interval.setter
    def update_interval(self, value):
        self._update_interval = value



#!/usr/bin/env python

"""
Local Processing Unit (LPU) with plugin support for various neuron/synapse models.
"""
import time
import collections
import pickle
import numbers
import copy
import itertools
import functools
import inspect, glob, os, importlib, sys

from future.utils import iteritems
from past.builtins import long
from builtins import zip


import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule
import pycuda.elementwise as elementwise

import numpy as np
import networkx as nx

#import time

# Work around bug in networkx < 1.9 that causes networkx to choke on GEXF
# files with boolean attributes that contain the strings 'True' or 'False'
# (bug already observed in https://github.com/networkx/networkx/pull/971)
nx.readwrite.gexf.GEXF.convert_bool['false'] = False
nx.readwrite.gexf.GEXF.convert_bool['False'] = False
nx.readwrite.gexf.GEXF.convert_bool['true'] = True
nx.readwrite.gexf.GEXF.convert_bool['True'] = True

from neurokernel.mixins import LoggerMixin
from neurokernel.core_gpu import Module, CTRL_TAG, GPOT_TAG, SPIKE_TAG
from neurokernel.tools.gpu import get_by_inds
from neurokernel.plsel import Selector

from types import *
from collections import Counter

from .utils.simpleio import *
from .utils import parray

from . import NDComponents as nd
from .NDComponents.NDComponent import NDComponent
from .MemoryManager import MemoryManager

import pdb

PORT_IN_GPOT = 'port_in_gpot'
PORT_IN_SPK = 'port_in_spk'
PORT_OUT_GPOT = 'port_out_gpot'
PORT_OUT_SPK = 'port_out_spk'

class LPU(Module):
    """
    Create an LPU circuit to simulate

    Parameters
    ----------
    dt: float
        simulation time step
    graph_type: - str
                  specify the method to generate/load the graph
                  options:
                  "obj": networkx.Graph object.
                  "gpickle": a networkx.Graph object pickled into a file by networkx.write_gpickle.
                  "pickle": a networkx.Graph object picked into a byte stream.
                  "gexf": a graph specified by a GEXF file, e.g. by networkx.write_gexf.
                  "function": a functools.partial object that have all appropriate arguments supplied.
                - dict (only for backward compatibility)
                  dictionary of components (nodes in a graph) and their properties
                  the dictionary is generated by `graph_to_dicts` method
    graph_dict: - dict
                  A dictionary specifying how to get/create the graph.
                  Depending on the value of graph_type, the following is required:
                  if graph_type is "obj" - "graph": networkx.Graph
                  if graph_type is "gpickle" - "filename": str
                  if graph_type is "pickle" - "stream": pickled byte object
                  if graph_type is "gexf" - "filename": str
                  if graph_type is "function" - "function": functools.partial object
                  The following is optional:
                  "kwargs": dict - keyword arguments for calling graph_to_dicts
                - str or networkx.Graph or byte or functools.partial
                  Depdending on the graph_type value, the corresponding
                  filename or object can be used
                - list (only for backward compatibility)
                  a list of tuples (pre, post, conn) specifying the edges of
                  a graph and their properties
    device: int
            specifies the device number to use (default: 0)
    input_processors: list
                      a list of InputProcessor objects (default: empty list)
    output_processors: list
                      a list of InputProcessor objects (default: empty list)
    ctrl_tag, gpot_tag, spike_tag: int
        MPI tags that respectively identify messages containing control data,
        graded potential port values, and spiking port values transmitted to
        worker nodes. (see neurokernel.core_gpu.Module)
    manager: bool
        Managerless running mode flag. It False, run Module without a
        manager. (default: True).
    rank_to_id: bidict.bidict
        Mapping between MPI ranks and module object IDs (default: None).
    routing_table: neurokernel.routing_table.RoutingTable
        Routing table describing data connections between modules. If no routing
        table is specified, the module will be executed in isolation. (default: None)
    uid_key: str
             key/property name in the graph that specifies the uid of nodes
             (default: 'id')
    debug: boolean
           run LPU in debug mode (default: False)
    columns: list of str
             Interface port attributes.
             Network port for controlling the module instance.
             (default: ['io', 'type', 'interface'])
    cuda_verbose: boolean
                  print out cuda kernel compilation message (defautl: False)
    time_sync: boolean
               count the time spent on communication (default: False)
    default_dtype: np.dtype
                   default numpy dtype to use for simulation (default: np.double)
    control_inteface: not implemented
    id: str
        Module identifier. If no identifier is specified, a unique
        identifier is automatically generated.
    extra_comps: NDComponent class or module or str or list
                 Provide the definition of NDComponent models to use in simulation.
                 NDComponent class: use the class directly
                 module: search for all NDComponent subclasses within module.
                         if module does not contain any NDComponent subclass
                         and is __init__.py, then walk through the folder and
                         its subfolders to retrieve all NDComponent classes
                         from all modules.
                 str: can be a path to a module or path to a directory.
                      If the former, import all NDComponent subclasses within module.
                      If the latter, import all NDComponent subclasses in modules
                      in the folder and any subfolders.
                 list: any combination of the above types.
                 (default: [])
    print_timing: boolean
                  to print out a detailed timing break out for quick profiling
                  (default: False)
    """
    @staticmethod
    def conv_legacy_graph(g):
        """
        Converts a gexf from legacy neurodriver format to one currently
        supported
        """


        # Find maximum ID in given graph so that we can use it to create new nodes
        # with IDs that don't overlap with those that already exist:
        max_id = 0
        for id in g.nodes():
            if isinstance(id, basestring):
                if id.isdigit():
                    max_id = max(max_id, int(id))
                else:
                    raise ValueError('node id must be an integer')
            elif isinstance(id, numbers.Integral):
                max_id = max(max_id, id)
            else:
                raise ValueError('node id must be an integer')
            gen_new_id = itertools.count(max_id+1).next

        # Create LPU and interface nodes and connect the latter to the former via an
        # Owns edge:
        g_new = nx.MultiDiGraph()

        # Transformation:
        # 1. nonpublic neuron node -> neuron node
        # 2. public neuron node -> neuron node with
        #    output edge to output port
        # 3. input port -> input port
        # 4. synapse edge -> synapse node + 2 edges connecting
        #    transformed original input/output nodes
        edges_to_out_ports = [] # edges to new output port nodes:
        for id, data in g.nodes(data=True):

            # Don't clobber the original graph's data:
            data = copy.deepcopy(data)

            if 'public' in data and data['public']:
                new_id = gen_new_id()
                port_data = {'selector': data['selector'],
                             'port_type': 'spike' if data['spiking'] else 'gpot',
                             'port_io': 'out',
                             'class': 'Port'}
                g_new.add_node(new_id, **port_data)
                edges_to_out_ports.append((id, new_id))
                del data['selector']

            if 'model' in data:
                if data['model'] == 'port_in_gpot':
                    for a in data:
                        if a!='selector': del data[a]
                    data['class'] = 'Port'
                    data['port_type'] = 'gpot'
                    data['port_io'] = 'in'
                elif data['model'] == 'port_in_spk':
                    for a in data:
                        if a!='selector': del data[a]
                    data['class'] = 'Port'
                    data['port_type'] = 'spike'
                    data['port_io'] = 'in'
                else:
                    data['class'] = data['model']

                # Don't need to several attributes that are implicit:
                for a in ['model', 'public', 'spiking','extern']:
                    if a in data: del data[a]

                g_new.add_node(id, **data)

        # Create synapse nodes for each edge in original graph and connect them to
        # the source/dest neuron/port nodes:
        for from_id, to_id, data in g.edges(data=True):
            data = copy.deepcopy(data)
            if data['model'] == 'power_gpot_gpot':
                data['class'] = 'PowerGPotGPot'
            else:
                data['class'] = data['model']
            del data['model']

            if 'id' in data: del data['id']

            new_id = gen_new_id()
            g_new.add_node(new_id, **data)
            g_new.add_edge(from_id, new_id)
            g_new.add_edge(new_id, to_id)

        # Connect output ports to the neurons that emit data through them:
        for from_id, to_id in edges_to_out_ports:
            g_new.add_edge(from_id, to_id)

        return g_new

    @staticmethod
    def graph_to_dicts(graph, uid_key=None, class_key='class',
                       remove_edge_id = True):
        """
        Convert graph of LPU neuron/synapse data to Python data structures.

        Parameters
        ----------
        graph : networkx.MultiDiGraph
            NetworkX graph containing LPU data.

        Returns
        -------
        comp_dict : dict
            A dictionary of components of which
            keys are model names, and
            values are dictionaries of parameters/attributes associated
            with the model.
            Keys of a dictionary of parameters are the names of them,
            and values of corresponding keys are lists of value of parameters.
            One of the parameters is called 'id' and by default it
            uses the id of the node in the graph.
            If uid_key is specified, id will use the specified parameter.
            Therefore, comp_dict has the following structure:

            comp_dict = {}
                comp_dict[model_name_1] = {}
                    comp_dict[model_name_1][parameter_1] = []
                    ...
                    comp_dict[model_name_1][parameter_N] = []
                    comp_dict[model_name_1][id] = []

                ...

                comp_dict[model_name_M] = {}
                    comp_dict[model_name_M][parameter_1] = []
                    ...
                    comp_dict[model_name_M][parameter_N] = []
                    comp_dict[model_name_M][id] = []

        conns : list
            A list of edges contained in graph describing the relation
            between components

        Example
        -------
        TODO: Update

        Notes
        -----
        TODO: Update
        """

        comp_dict = {}
        comps = graph.nodes.items()

        all_component_types = list(set([comp[class_key] for uid, comp in comps]))

        for model in all_component_types:
            sub_comps = [comp for comp in comps \
                                   if comp[1][class_key] == model]

            all_keys = [set(comp[1]) for comp in sub_comps]
            key_intersection = set.intersection(*all_keys)
            key_union = set.union(*all_keys)

            # For visually checking if any essential parameter is dropped
            ignored_keys = list(key_union-key_intersection)
            if ignored_keys:
                print('parameters of model {} ignored: {}'.format(model, ignored_keys))

            del all_keys

            sub_comp_keys = list(key_intersection)

            if model == 'Port':
                assert('selector' in sub_comp_keys)

            comp_dict[model] = {
                k: [comp[k] for uid, comp in sub_comps] \
                for k in sub_comp_keys if not k in [uid_key, class_key]}

            comp_dict[model]['id'] = [comp[uid_key] if uid_key else uid \
                                      for uid, comp in sub_comps]

#        for id, comp in comps:
#            model = comp[class_key]
#
#            # For port, make sure selector is specified
#            if model == 'Port':
#                assert('selector' in comp.keys())
#
#            # if the neuron model does not appear before, add it into n_dict
#            if model not in comp_dict:
#                comp_dict[model] = {k:[] for k in comp.keys() + ['id']}
#
#            # Same model should have the same attributes
#            if not set(comp_dict[model].keys()) == set(comp.keys() + ['id']):
#                raise KeyError("keys of component does not match with that of "+\
#                               model+": "+ str(set(comp_dict[model].keys())) +
#                               str(set(comp.keys() + ['id'])))
#
#            # add data to the subdictionary of comp_dict
#            for key in comp.iterkeys():
#                if not key==uid_key:
#                    comp_dict[model][key].append( comp[key] )
#            if uid_key:
#                comp_dict[model]['id'].append(comp[uid_key])
#            else:
#                comp_dict[model]['id'].append( id )
#
#        # Remove duplicate model information:
#        for val in comp_dict.itervalues(): val.pop(class_key)

        # Extract connections
        conns = graph.edges(data=True)
        if remove_edge_id:
            for pre, post, conn in conns:
                conn.pop('id', None)
                conn.pop('tag', None)
        return comp_dict, conns

    @staticmethod
    def lpu_parser(filename):
        """
        GEXF LPU specification parser.

        Extract LPU specification data from a GEXF file and store it in
        Python data structures.
        TODO: Update

        Parameters
        ----------
        filename : str
            GEXF filename.

        Returns
        -------
        TODO: Update
        """

        graph = nx.read_gexf(filename)
        return LPU.graph_to_dicts(graph, remove_edge_id = True)

    @staticmethod
    def lpu_parser_legacy(filename):
        """
        TODO: Update
        """

        graph = nx.read_gexf(filename)
        return LPU.graph_to_dicts(LPU.conv_legacy_graph(graph))

    @classmethod
    def extract_in_gpot(cls, comp_dict, uid_key):
        """
        Return selectors of non-spiking input ports.
        """
        if not 'Port' in comp_dict: return ('',[])
        a = list(zip(*[(sel,uid) for sel,ptype,io,uid in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_type'],
                             comp_dict['Port']['port_io'],
                             comp_dict['Port'][uid_key]) if ptype=='gpot' \
                  and io=='in']))
        if not a: a = ('',[])
        return a

    @classmethod
    def extract_in_spk(cls, comp_dict, uid_key):
        """
        Return selectors of spiking input ports.
        """
        if not 'Port' in comp_dict: return ('',[])
        a = list(zip(*[(sel,uid) for sel,ptype,io,uid in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_type'],
                             comp_dict['Port']['port_io'],
                             comp_dict['Port'][uid_key]) if ptype=='spike' \
                         and io=='in']))
        if not a: a = ('',[])
        return a

    @classmethod
    def extract_out_gpot(cls, comp_dict, uid_key):
        """
        Return selectors of non-spiking output neurons.
        """
        if not 'Port' in comp_dict: return ('',[])
        a = list(zip(*[(sel,uid) for sel,ptype,io,uid in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_type'],
                             comp_dict['Port']['port_io'],
                             comp_dict['Port'][uid_key]) if ptype=='gpot' \
                  and io=='out']))
        if not a: a = ('',[])
        return a

    @classmethod
    def extract_out_spk(cls, comp_dict, uid_key):
        """
        Return selectors of spiking output neurons.
        """
        if not 'Port' in comp_dict: return ('',[])
        a = list(zip(*[(sel,uid) for sel,ptype,io,uid in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_type'],
                             comp_dict['Port']['port_io'],
                             comp_dict['Port'][uid_key]) if ptype=='spike' \
                  and io=='out']))
        if not a: a = ('',[])
        return a

    @classmethod
    def extract_sel_in_gpot(cls, comp_dict):
        """
        Return selectors of non-spiking input ports.
        """
        if not 'Port' in comp_dict: return ''
        return ','.join([sel  for sel,ptype,io in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_type'],
                             comp_dict['Port']['port_io']) \
                             if ptype=='gpot' and io=='in'])

    @classmethod
    def extract_sel_in_spk(cls, comp_dict):
        """
        Return selectors of spiking input ports.
        """
        if not 'Port' in comp_dict: return ''
        return ','.join([sel  for sel,ptype,io in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_type'],
                             comp_dict['Port']['port_io']) \
                             if ptype=='spike' and io=='in'])

    @classmethod
    def extract_sel_out_gpot(cls, comp_dict):
        """
        Return selectors of non-spiking output neurons.
        """
        if not 'Port' in comp_dict: return ''
        return ','.join([sel  for sel,ptype,io in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_type'],
                             comp_dict['Port']['port_io']) \
                             if ptype=='gpot' and io=='out'])

    @classmethod
    def extract_sel_out_spk(cls, comp_dict):
        """
        Return selectors of spiking output neurons.
        """
        if not 'Port' in comp_dict: return ''
        return ','.join([sel for sel,ptype,io,uid in \
                  zip(comp_dict['Port']['selector'],
                      comp_dict['Port']['port_type'],
                      comp_dict['Port']['port_io']) \
                  if ptype=='spike' and io=='out'])

    @classmethod
    def extract_sel_in(cls, comp_dict):
        """
        Return selectors of all input ports.
        """
        if not 'Port' in comp_dict: return ''
        return ','.join([sel for sel, io in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_io']) if io=='in'])

    @classmethod
    def extract_sel_out(cls, comp_dict):
        """
        Return selectors of all output neurons.
        """

        if not 'Port' in comp_dict: return ''
        return ','.join([sel for sel, io in \
                         zip(comp_dict['Port']['selector'],
                             comp_dict['Port']['port_io']) if io=='out'])

    @classmethod
    def extract_sel_all(cls, comp_dict):
        """
        Return selectors for all input ports and output neurons.
        """

        return ','.join(filter(None, \
                    [cls.extract_in(comp_dict), cls.extract_out(comp_dict)]))

    @classmethod
    def import_NDcomponent_from_path(cls, path):
        cls = []
        if not isinstance(path, str):
            raise TypeError('path must be a str')

        if os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                if os.path.split(root)[-1][0] in ['_','.']:
                    continue
                sys.path.append(root)
                for file in files:
                    m, ext = os.path.splitext(file)
                    if ext == '.py' and not file.startswith('_'):
                        try:
                            mod = importlib.import_module(m, root)
                            cls.extend(inspect.getmembers(
                                        mod,
                                        lambda x: inspect.isclass(x) and \
                                        issubclass(x, NDComponent) and \
                                        x.__module__ == mod.__name__))
                        except ImportError:
                            print('failed to import from {} {}'.format(root, m))
                            pass
                sys.path.pop(-1)
        elif os.path.isfile(path):
            p, file = os.path.split(path)
            m, ext = os.path.splitext(file)
            if ext == '.py' and not file.startswith('_'):
                sys.path.append(p)
                try:
                    mod = importlib.import_module(m, p)
                    cls.extend(inspect.getmembers(
                                mod,
                                lambda x: inspect.isclass(x) and \
                                          issubclass(x, NDComponent) and \
                                          x.__module__ == mod.__name__))
                except ImportError:
                    print('failed to import from {} {}'.format(p, m))
                    pass
                sys.path.pop(-1)
        else:
            raise FileNotFoundError('File or directory does not exist: {}'.format(path))
        return cls

    @classmethod
    def import_NDcomponent_from_module(cls, mod):
        if not inspect.ismodule(mod):
            raise TypeError('mod is not a module')
        cls = inspect.getmembers(
                    mod,
                    lambda x: inspect.isclass(x) and issubclass(x, NDComponent))
        if not len(cls):
            # If module does not contain any NDComponent subclasses,
            # check if the module is an __init__.py
            # if it is, then recursively import all importable NDComponent classes
            # down the folder and subfolders
            path, file = os.path.split(mod.__file__)
            if file == '__init__.py':
                base_package = mod.__package__
                to_remove = len(path.split('/'))
                for root, dirs, files in os.walk(path):
                    if os.path.split(root)[-1][0] in ['_', '.']:
                        continue
                    subpackage = '.'.join(['']+root.split('/')[to_remove:])
                    for file in files:
                        m, ext = os.path.splitext(file)
                        if ext == '.py' and not file.startswith('_'):
                            try:
                                mod = importlib.import_module(
                                        '{}.{}'.format(subpackage, m),
                                        base_package)
                                cls.extend(inspect.getmembers(
                                            mod,
                                            lambda x: inspect.isclass(x) and \
                                            issubclass(x, NDComponent) and \
                                            x.__module__ == mod.__name__))
                            except ImportError:
                                print('failed to import from {} {}'.format(m, subpackage))
                                pass
        return cls

    def import_models(self, models):
        if inspect.isclass(models) and issubclass(models, NDComponent):
            # When an NDComponent class is imported directly
            self._comps[models.__name__] = {'accesses': models.accesses ,
                                            'updates':models.updates,
                                            'cls':models}
        elif inspect.ismodule(models):
            # when a module is provided
            cls = self.import_NDcomponent_from_module(models)
            for name, obj in cls:
                self._comps[name] = {'accesses': obj.accesses,
                                     'updates':obj.updates,
                                     'cls':obj}
        elif isinstance(models, str):
            cls = self.import_NDcomponent_from_path(models)
            for name, obj in cls:
                self._comps[name] = {'accesses': obj.accesses,
                                     'updates':obj.updates,
                                     'cls':obj}
        elif isinstance(models, (list, tuple)):
            for model in models:
                self.import_models(model)


    def __init__(self, dt, graph_type, graph_dict, device=0, input_processors=[],
                 output_processors=[], ctrl_tag=CTRL_TAG, gpot_tag=GPOT_TAG,
                 spike_tag=SPIKE_TAG, manager = True,
                 rank_to_id=None, routing_table=None,
                 uid_key='id', debug=False, columns=['io', 'type', 'interface'],
                 cuda_verbose=False, time_sync=False, default_dtype=np.double,
                 control_inteface=None, id=None, extra_comps=[],
                 print_timing=False):
        compilation_start_time = time.time()

        LoggerMixin.__init__(self, 'LPU {}'.format(id))

        assert('io' in columns)
        assert('type' in columns)
        assert('interface' in columns)
        self.LPU_id = id
        self.dt = dt
        self.step = 0
        self._debug = debug
        self.device = device
        self._default_dtype = default_dtype
        self.control_inteface = control_inteface
        self._print_timing = print_timing
        if cuda_verbose:
            self._compile_options = ['--ptxas-options=-v']
        else:
            self._compile_options = []

        if self._print_timing:
            start = time.time()

        if not isinstance(input_processors, list):
            input_processors = [input_processors]
        if not isinstance(output_processors, list):
            output_processors = [output_processors]

        self.output_processors = output_processors
        self.input_processors = input_processors

        G = nx.MultiDiGraph()
        for input_processor in self.input_processors:
            G = nx.compose(G, input_processor.get_new_graph())

        if isinstance(graph_type, dict):
        # for backward compatibility where the two arguments are comp_dict and conn_list
            comp_dict = graph_type
            conn_list = graph_dict
        elif isinstance(graph_type, str):
            if graph_type == 'obj':
                # use networkx.Graph objects,
                # as slow as using comp_dict, conn_list directly
                if isinstance(graph_dict, nx.Graph):
                    kwargs = {}
                    graph = graph_dict
                elif isinstance(graph_dict, dict):
                    kwargs = graph_dict.get('kwargs', {})
                    graph = graph_dict['graph']
                else:
                    raise TypeError("For graph_type = 'obj', graph must be networkx Graphs")
            elif graph_type == 'gpickle':
                # use networkx built-in pickle function,
                # relatively fast but require disk space
                if isinstance(graph_dict, str):
                    kwargs = {}
                    graph = nx.readwrite.gpickle.read_gpickle(graph_dict)
                elif isinstance(graph_dict, dict):
                    kwargs = graph_dict.get('kwargs', {})
                    graph = nx.read_gpickle(graph_dict['filename'])
                else:
                    raise TypeError('graph_dict must be either str (filename) or dict')
            elif graph_type == 'pickle':
                # use pickled byte stream of networkx.Graph
                # fast
                if isinstance(graph_dict, bytes):
                    kwargs = {}
                    graph = pickle.loads(graph_dict)
                elif isinstance(graph_dict, dict):
                    kwargs = graph_dict.get('kwargs', {})
                    graph = pickle.loads(graph_dict['stream'])
                else:
                    raise TypeError('graph_dict must be either pickled byte stream or dict')
            elif graph_type == 'gexf':
                # use networkx build-in read_gexf
                # slow and require disk space
                if isinstance(graph_dict, str):
                    kwargs = {'remove_edge_id': True}
                    graph = nx.read_gexf(graph_dict)
                elif isinstance(graph_dict, dict):
                    kwargs = graph_dict.get('kwargs', {'remove_edge_id': True})
                    graph = nx.read_gexf(graph_dict['filename'])
                else:
                    raise TypeError('graph_dict must be either str (filename) or dict')
            elif graph_type == 'function':
                # call a function created by functools.partial method,
                # where the function and arguments
                # fast and generate the graph during initialization
                if isinstance(graph_dict, (functools.partial, functools.partialmethod)):
                    kwargs = {}
                    graph = graph_dict()
                elif isinstance(graph_dict, dict):
                    kwargs = graph_dict.get('kwargs', {})
                    graph = graph_dict['function']()
            else:
                raise TypeError('Unkown method to generate Graph')
            comp_dict, conn_list = self.graph_to_dicts(nx.compose(graph, G), **kwargs)
        else:
            raise TypeError('Unkown method to generate Graph')

        if self._print_timing:
            self.log_info("Elapsed time for converting graph: {:.3f} seconds".format(time.time()-start))
            start = time.time()

        self._uid_key = uid_key

        self._uid_generator = uid_generator()

        # Load all NDComponents:
        self._comps = {}
        self.import_models(nd)
        if extra_comps:
            self.import_models(extra_comps)

        if self._print_timing:
            self.log_info("Elapsed time for importing component definitions: {:.3f} seconds".format(time.time()-start))
            start = time.time()

        # self._comps = {cls.__name__:{'accesses': cls.accesses ,
        #                              'updates':cls.updates,
        #                              'cls':cls} \
        #                for cls in comp_classes if not cls.__name__[:4]=='Base'}
        # self._load_components(extra_comps=extra_comps)

        # Ignore models without implementation
        models_to_be_deleted = []
        for model in comp_dict:
            if not model in self._comps and not model in ['Port','Input']:
                self.log_info("Ignoring Model %s: Can not find implementation"
                              % model)
                raise NotImplementedError("Model %s: Can not find implementation" % model)
                models_to_be_deleted.append(model)
        for model in models_to_be_deleted:
            del comp_dict[model]

        # Assume zero delay by default
        self._variable_delay_map = {}

        # Generate a uid to model map of components
        self._uid_model_map = {}
        for model,attribs in iteritems(comp_dict):
            for i,uid in enumerate(attribs[uid_key]):
                self._uid_model_map[uid] = model


        # Map from post synaptic component to aggregator uid
        agg_map = {}
        agg = {}

        #start = time.time()

        conns = []
        self._in_port_vars = {}
        self._out_port_conns = []
        comp_uid_order = {}
        for model, attribs in comp_dict.items():
            comp_uid_order[model] = {uid: i for i, uid in enumerate(attribs[uid_key])}

        if self._print_timing:
            self.log_info("Elapsed time for processing comp_dict: {:.3f} seconds".format(time.time()-start))
            start = time.time()
        # Process connections between components, remove inconsitent connections
        # calculate required delays, infer variable if required
        in_port_vars_set = {}
        reverse_set = ['reverse','Vr','VR','reverse_potential']

        comp_updates = {name: v['updates'] if not name=='Port' else [] \
                        for name, v in self._comps.items()}
        comp_updates.update({'Port': []})
        comp_accesses = {name: v['accesses'] if not name=='Port' else [] \
                         for name, v in self._comps.items()}
        comp_accesses.update({'Port': []})

        for pre, post, data in conn_list:
            try:
                pre_model = self._uid_model_map[pre]
            except KeyError:
                continue
            try:
                post_model = self._uid_model_map[post]
            except KeyError:
                continue

            pre_updates = comp_updates[pre_model]
            pre_updates_set = set(pre_updates)
            post_accesses = comp_accesses[post_model]
            post_accesses_set = set(post_accesses)
            pre_post_intersection = pre_updates_set&post_accesses_set
            g_in_pre_update = 'g' in pre_updates_set # not so useful
            V_in_pre_update = 'V' in pre_updates_set

            # Update delay
            delay = max(int(round((data['delay']/dt))) \
                    if 'delay' in data else 0, 1) - 1
            data['delay'] = delay

            if post_model == 'Aggregator':
                agg_map[post] = post
                reverse_key = None
                for k in reverse_set:
                    if k in data:
                        reverse_key = k
                        reverse = data[reverse_key]
                        break
                if reverse_key is None:
                    # else look in the attibutes of the synapse
                    s = (set(['reverse','Vr','VR','reverse_potential'])&
                         set(comp_dict[pre_model]))
                    if s: reverse_key = s.pop()
                    if reverse_key:
                        reverse = comp_dict[pre_model][reverse_key][\
                                comp_uid_order[pre_model][pre]]
                        if g_in_pre_update:
                            data['reverse'] = reverse
                    else:
                        if g_in_pre_update:
                            self.log_info('Assuming reverse potential ' +
                                          'to be zero for connection from' +
                                          '%s to %s'%(pre,post))
                            data['reverse'] = 0
                        reverse = 0

                if post in agg:
                    if g_in_pre_update:
                        agg[post].append({'pre':pre,'reverse':reverse,
                                          'variable':'g'})
                    elif V_in_pre_update:
                        agg[post].append({'pre':pre, 'variable':'V'})
                else:
                    # Make sure aggregator has access to postsynaptic voltage
                    if g_in_pre_update:
                        agg[post] = [{'pre':pre,'reverse':reverse,'variable':'g'}]
                    elif V_in_pre_update:
                        agg[post] = [{'pre':pre,'variable':'V'}]

                if g_in_pre_update:
                    agg[post][-1].update(data)
                    self._variable_delay_map['g'] = max(data['delay'],
                                    self._variable_delay_map['g'] if 'g' in \
                                    self._variable_delay_map else 0)


            # Ensure consistency
            # Insert Aggregator between g->V if required. Assume 'reverse' or
            # 'Vr' or 'VR' or 'reverse_potential' id present as a param in the
            # synapse in that case
            if not pre_post_intersection:
                if g_in_pre_update and 'I' in post_accesses_set:
                    # start2 = time.time()
                    # First look for reverse in the attributes of the edge
                    reverse_key = None
                    for k in reverse_set:
                        if k in data:
                            reverse_key = k
                            reverse = data[reverse_key]
                            break
                    if reverse_key is None:
                        # else look in the attibutes of the synapse
                        s = (set(['reverse','Vr','VR','reverse_potential'])&
                             set(comp_dict[pre_model]))
                        if s: reverse_key = s.pop()
                        if reverse_key:
                            reverse = comp_dict[pre_model][reverse_key][\
                                    comp_uid_order[pre_model][pre]]
                        else:
                            self.log_info('Assuming reverse potential ' +
                                          'to be zero for connection from' +
                                          '%s to %s'%(pre,post))
                            reverse = 0

                    data.update({'pre':pre,'reverse':reverse,
                                 'variable':'g'})
                    if post in agg:
                        agg[post].append(data)
                    else:
                        # Make sure aggregator has access to postsynaptic voltage
                        agg[post] = [{'pre':post,'variable':'V'},
                                     data]
                    if post not in agg_map:
                        uid = self._uid_generator.generate_uid()
                        agg_map[post] = uid
                    self._variable_delay_map['g'] = max(data['delay'],
                                    self._variable_delay_map['g'] if 'g' in \
                                    self._variable_delay_map else 0)

                elif pre_model == 'Port':
                    if not 'variable' in data:
                        data['variable'] = post_accesses[0]
                    if not data['variable'] in self._in_port_vars:
                        self._in_port_vars[data['variable']] = []
                        in_port_vars_set[data['variable']] = set()
                    if pre not in in_port_vars_set[data['variable']]:
                        self._in_port_vars[data['variable']].append(pre)
                        in_port_vars_set[data['variable']].add(pre)
                    conns.append((pre, post, data))
                    self._variable_delay_map[data['variable']] = max(data['delay'],
                            self._variable_delay_map[data['variable']] if \
                            data['variable'] in self._variable_delay_map else 0)
                elif post_model == 'Port':
                    if not 'variable' in data:
                        data['variable'] = pre_updates[0]
                    self._out_port_conns.append((pre, post, data['variable']))
                else:
                    self.log_info("Ignoring connection %s -> %s"%(pre,post))
                continue

            var = data['variable'] if 'variable' in data else None
            if not var:
                var = pre_post_intersection.pop()
            elif not (var in pre_updates_set and var in post_accesses_set):
                continue
            data['variable'] = var
            self._variable_delay_map[data['variable']] = max(data['delay'],
                            self._variable_delay_map[data['variable']] if \
                            data['variable'] in self._variable_delay_map else 0)
            # connection to Aggregator will be added later
            if 'Aggregator' not in post_model:
                conns.append((pre,post,data))

        if self._print_timing:
            self.log_info("Elapsed time for processing conn_list: {:.3f} seconds".format(time.time()-start))
            start = time.time()

        if agg and not 'Aggregator' in comp_dict:
            comp_dict['Aggregator'] = {uid_key: []}
        if agg:
            agg_uid_key_set = set(comp_dict['Aggregator'][uid_key])

        # Add updated aggregator components to component dictionary
        # and create connections for aggregator
        for post, conn_list in agg.items():
            uid = agg_map[post]
            if uid not in agg_uid_key_set:
                keys = [k for k in comp_dict['Aggregator'] if k != uid_key]
                comp_dict['Aggregator'][uid_key].append(uid)
                agg_uid_key_set.add(uid)
                self._uid_model_map[uid] = 'Aggregator'
                for k in keys:
                    comp_dict['Aggregator'][k].append(str(uid))
            for conn in conn_list:
                conns.append((conn['pre'],uid,{k:v for k,v in conn.items()
                                               if k!='pre'}))
            # Add a 'I' connection between Aggregator and neuron if they are
            # automatically generated.
            # This can be checking if the 'pre' attribute in the item
            # in conn_list with 'variable' 'V' is the same neuron as post
            if post == [tmp['pre'] for tmp in conn_list if tmp['variable']=='V'][0]:
                conns.append((uid,post,{'variable':'I', 'delay': 0}))
        if self._print_timing:
            self.log_info("Elapsed time for processing aggregator: {:.3f} seconds".format(time.time()-start))
            start = time.time()

        self.conn_dict = {}

        # RePackage connections
        for (pre, post, data) in conns:
            if not post in self.conn_dict:
                self.conn_dict[post] = {}
            var = data['variable']
            data.pop('variable')
            if not var in self.conn_dict[post]:
                self.conn_dict[post][var] = {k:[] for k in ['pre'] + list(data)}
            self.conn_dict[post][var]['pre'].append(pre)
            for k in data: self.conn_dict[post][var][k].append(data[k])

        if self._print_timing:
            self.log_info("Elapsed time for repackaging connections: {:.3f} seconds".format(time.time()-start))
            start = time.time()

        # Add connections for component with no incoming connections
        for uid, model in iteritems(self._uid_model_map):
            if model == 'Port':
                continue
            for var in self._comps[model]['accesses']:
                if ((not uid in self.conn_dict or not var in self.conn_dict[uid])):
                    pre = self._uid_generator.generate_uid(input=True)
                    if not var in self._variable_delay_map:
                        self._variable_delay_map[var]=0
                    if not uid in self.conn_dict: self.conn_dict[uid] = {}
                    if model == 'Aggregator' and var == 'g':
                        self.conn_dict[uid][var] = {'pre':[pre],'delay':[0],
                                                    'reverse': [0]} #'id': [0],
                    else:
                        self.conn_dict[uid][var] = {'pre':[pre],'delay':[0]}
                    if not 'Input' in comp_dict:
                        comp_dict['Input'] = {}
                    if not var in comp_dict['Input']:
                        comp_dict['Input'][var] = {self._uid_key: []}
                    comp_dict['Input'][var][self._uid_key].append(pre)

        if self._print_timing:
            self.log_info("Elapsed time for adding connections for component with no incoming connections: {:.3f} seconds".format(time.time()-start))
            start = time.time()

        # Optimize ordering (TODO)
        self._uid_ind_map = {m:{uid:i for i,uid in enumerate(n[uid_key])}
                            for m,n in comp_dict.items() if not m=='Input'}

        if 'Input' in comp_dict:
            self._uid_ind_map['Input'] = {var:{uid:i for i, uid in enumerate(d[uid_key])}
                                         for var, d in comp_dict['Input'].items()}

        # Reorder components
        for m, n in comp_dict.items():
            if m=='Input':
                for var, d in n.items():
                    order = np.argsort([self._uid_ind_map[m][var][uid] for uid in d[uid_key]])
                    d[uid_key] = [d[uid_key][i] for i in order]
                continue

            order = np.argsort([self._uid_ind_map[m][uid] for uid in n[uid_key]])
            for k in n:
                n[k] = [n[k][i] for i in order]

        # Reorder input port variables
        for var, uids in self._in_port_vars.items():
            order = np.argsort([self._uid_ind_map['Port'][uid] for uid in uids])
            self._in_port_vars[var] = [uids[i] for i in order]

        if self._print_timing:
            self.log_info("Elapsed time for optimizing ordering: {:.3f} seconds".format(time.time()-start))
            start = time.time()

        # Try to figure out order of stepping through components
        # If a loop of dependencies is present, update order behaviour is undefined
        models = list(comp_dict)
        try:
            models.remove('Port')
        except:
            pass

        try:
            models.remove('Input')
        except:
            pass


        deps = {i:[] for i in range(len(models))}
        for i in range(len(models)):
            for j in range(i+1,len(models)):
                in12 = set(self._comps[models[i]]['updates'])&\
                       set(self._comps[models[j]]['accesses'])
                in21 = set(self._comps[models[i]]['accesses'])&\
                       set(self._comps[models[j]]['updates'])
                if in12 or in21:
                    if len(in12) > len(in21):
                        deps[j].append(i)
                    else:
                        deps[i].append(j)


        self._exec_order = []
        for i, model in enumerate(models):
            if not model in self._exec_order: self._exec_order.append(model)
            for j in deps[i]:
                try:
                    if self._exec_order.index(models[j]) > \
                       self._exec_order.index(model):
                        self._exec_order.remove(models[j])
                        self._exec_order.insert(self._exec_order.index(model),
                                               models[j])
                except ValueError:
                    self._exec_order.insert(self._exec_order.index(model),
                                           models[j])

        var_mod = {}
        for i, model in enumerate(models):
            for var in self._comps[model]['updates']:
                if not var in var_mod: var_mod[var] = []
                var_mod[var].append(model)

        self._model_var_inj = {}
        for var, models in var_mod.items():
            i = 0
            for model in models:
                i = max(self._exec_order.index(model),i)
            if not self._exec_order[i] in self._model_var_inj:
                self._model_var_inj[self._exec_order[i]] = []
            self._model_var_inj[self._exec_order[i]].append(var)

        #Variables not updated by any component (for example those coming from
        #external input or Ports) are slated to be injected at the end of a step
        for var in self._variable_delay_map:
            if not var in var_mod:
                if not self._exec_order[-1] in self._model_var_inj:
                    self._model_var_inj[self._exec_order[-1]] = []
                self._model_var_inj[self._exec_order[-1]].append(var)

        if self._print_timing:
            self.log_info("Elapsed time for ordering execution: {:.3f} seconds".format(time.time()-start))
            start = time.time()
        # Get selectors of input ports:
        sel_in_gpot, self._in_gpot_uids = self.extract_in_gpot(comp_dict,
                                                              self._uid_key)
        self._sel_in_gpot = Selector(','.join(sel_in_gpot))
        sel_in_spk, self._in_spk_uids = self.extract_in_spk(comp_dict,
                                                           self._uid_key)
        self._sel_in_spk = Selector(','.join(sel_in_spk))
        sel_in = Selector.add(self._sel_in_gpot, self._sel_in_spk)

        # Get selectors of output neurons:
        sel_out_gpot, self._out_gpot_uids = self.extract_out_gpot(comp_dict,
                                                                 self._uid_key)
        self._sel_out_gpot = Selector(','.join(sel_out_gpot))
        sel_out_spk, self._out_spk_uids = self.extract_out_spk(comp_dict,
                                                              self._uid_key)
        self._sel_out_spk = Selector(','.join(sel_out_spk))
        sel_out = Selector.add(self._sel_out_gpot, self._sel_out_spk)
        sel_gpot = Selector.add(self._sel_in_gpot, self._sel_out_gpot)
        sel_spk = Selector.add(self._sel_in_spk, self._sel_out_spk)
        sel = Selector.add(sel_gpot, sel_spk)
        if self._print_timing:
            self.log_info("Elapsed time for generating selectors: {:.3f} seconds".format( time.time()-start))
            start = time.time()

        # Save component parameters data in the form
        # [('Model0', {'attrib0': [..], 'attrib1': [..]}), ('Model1', ...)]
        self.comp_list = comp_dict.items()
        self.models = {m:i for i,(m,_) in enumerate(self.comp_list)}

        # Number of components of each model:
        self.model_num = [len(n[uid_key]) if not m=='Input' else
                          len(sum([d[uid_key] for d in n.values()],[]))
                          for m, n in self.comp_list]

        data_gpot = np.zeros(len(self._in_gpot_uids)+len(self._out_gpot_uids),
                             self._default_dtype)
        data_spike = np.zeros(len(self._in_spk_uids)+len(self._out_spk_uids),
                              self._default_dtype)

        if self._print_timing:
            self.log_info("Elapsed time for misc: {:.3f} seconds".format(time.time()-start))
            start = time.time()
        super(LPU, self).__init__(sel=sel, sel_in=sel_in, sel_out=sel_out,
                                  sel_gpot=sel_gpot, sel_spike=sel_spk,
                                  data_gpot=data_gpot, data_spike=data_spike,
                                  columns=columns, ctrl_tag=ctrl_tag, gpot_tag=gpot_tag,
                                  spike_tag=spike_tag, id=self.LPU_id,
                                  manager = manager,
                                  rank_to_id=rank_to_id, routing_table=routing_table,
                                  device=device, debug=debug, time_sync=time_sync,
                                  print_timing=print_timing)

        for model in comp_dict:
            if model == 'Input':
                print('{}: Number of {}: {}'.format(self.id, model, {k: len(v[self._uid_key]) for k, v in comp_dict[model].items()}))
                self.log_info('Number of {}: {}'.format(model, {k: len(v[self._uid_key]) for k, v in comp_dict[model].items()}.__repr__().replace('{','').replace('}','')))
            else:
                print('{}: Number of {}: {}'.format(self.id, model, len(comp_dict[model][self._uid_key])))
                self.log_info('Number of {}: {}'.format(model, len(comp_dict[model][self._uid_key])))

        if self._print_timing:
            cuda.Context.synchronize()
            self.log_info("Elapsed time for initializing parent class: {:.3f} seconds".format(time.time()-start))


        # Integer indices in port map data arrays corresponding to input/output
        # gpot/spiking ports:
        self._in_gpot_inds = np.array(self.pm['gpot'].ports_to_inds(
                                     self._sel_in_gpot), dtype=np.int32)
        self._out_gpot_inds = np.array(self.pm['gpot'].ports_to_inds(
                                      self._sel_out_gpot), dtype=np.int32)
        self._in_spk_inds = np.array(self.pm['spike'].ports_to_inds(
                                    self._sel_in_spk), dtype=np.int32)
        self._out_spk_inds = np.array(self.pm['spike'].ports_to_inds(
                                     self._sel_out_spk), dtype=np.int32)

        self._compile_time = time.time()-compilation_start_time

    # def generate_uid(self, input=False):
    #     if input:
    #         uid = 'input_' + str(np.random.randint(100000))
    #     else:
    #         uid = 'auto_' + str(np.random.randint(100000))
    #     while uid in self.gen_uids:
    #         if input:
    #             uid = 'input_' + str(np.random.randint(100000))
    #         else:
    #             uid = 'auto_' + str(np.random.randint(100000))
    #     return uid

    def pre_run(self):
        pre_run_start_time = time.time()
        if self._print_timing:
            start = time.time()
        super(LPU, self).pre_run()
        if self._print_timing:
            self.log_info("LPU pre_run parent took {} seconds".format(time.time()-start))
            start = time.time()

        self.memory_manager = MemoryManager()
        self.init_variable_memory()
        if self._print_timing:
            cuda.Context.synchronize()
            self.log_info('Elapsed time for initialing variable memory: {:.3f} seconds'.format( time.time()-start))
            start = time.time()

        self.process_connections()
        if self._print_timing:
            self.log_info('Elapsed time for process_connections: {:.3f} seconds'.format(time.time()-start))
            start = time.time()

        self.init_parameters()
        if self._print_timing:
            cuda.Context.synchronize()
            self.log_info('Elapsed time for init_paramseters: {:.3f} seconds'.format( time.time()-start))
            start = time.time()

        self.components = {}
        # Instantiate components
        for model in self.models:
            if model in ['Port','Input']: continue
            self.components[model] = self._instantiate_component(model)
            update_pointers = {}
            for var in self._comps[model]['updates']:
                buff = self.memory_manager.get_buffer(var)
                mind = self.memory_manager.variables[var]['models'].index(model)
                shift = self.memory_manager.variables[var]['cumlen'][mind]
                update_pointers[var] = int(buff.gpudata)+(buff.current*buff.ld+\
                                            shift)*buff.dtype.itemsize
            self.components[model].pre_run(update_pointers)
            for var in self._comps[model]['updates']:
                buff = self.memory_manager.get_buffer(var)
                mind = self.memory_manager.variables[var]['models'].index(model)
                shift = self.memory_manager.variables[var]['cumlen'][mind]
                for j in range(buff.buffer_length):
                    if j is not buff.current:
                        cuda.memcpy_dtod(
                            int(int(buff.gpudata)+(j*buff.ld+\
                                                shift)*buff.dtype.itemsize),
                            int(int(buff.gpudata)+(buff.current*buff.ld+\
                                                shift)*buff.dtype.itemsize),
                            int(buff.dtype.itemsize*self.model_num[self.models[model]]))
        if self._print_timing:
            cuda.Context.synchronize()
            self.log_info('Elapsed time for instantiating components: {:.3f} seconds'.format(time.time()-start))
            start = time.time()
        # Setup ports
        self._setup_input_ports()
        self._setup_output_ports()

        if self._print_timing:
            cuda.Context.synchronize()
            self.log_info('Elapsed time for setting up ports: {:.3f} seconds'.format( time.time()-start))
            start = time.time()

        for p in self.input_processors:
            p.LPU_obj = self
            p._pre_run()

        for p in self.output_processors:
            p.LPU_obj = self
            p._pre_run()

        if self._print_timing:
            cuda.Context.synchronize()
            self.log_info('Elapsed time for prerun input and output processors: {:.3f} seconds'.format( time.time()-start))
            start = time.time()

        self.memory_manager.precompile_fill_zeros()

        if self.control_inteface: self.control_inteface.register(self)
        if self._print_timing:
            cuda.Context.synchronize()
            self.log_info("Elapsed time for LPU pre_run: {:.3f} seconds".format(time.time()-pre_run_start_time))

        self._compile_time += time.time()-pre_run_start_time

        if self._print_timing:
            self.timing = {'read_input': 0, 'input_processors': 0, 'inject_input': 0,
                           'model_run': 0, 'output_processors': 0,
                           'extract_output': 0, 'total': 0}
        cuda.Context.synchronize()
        self.log_info("Total compiling time: {:.3f} seconds".format(self._compile_time))
        print('Compilation of executable circuit completed in {} seconds'.format(
                self._compile_time))

    # TODO: optimize the order of self._out_port_conns beforehand
    def _setup_output_ports(self):
        self._out_port_inds_gpot = {}
        self._out_var_inds_gpot = {}
        self._out_port_inds_spk = {}
        self._out_var_inds_spk = {}
        # assuming that the UIDs are unique
        out_gpot_index = {uid: i for i, uid in enumerate(self._out_gpot_uids)}
        out_spk_index = {uid: i for i, uid in enumerate(self._out_spk_uids)}
        for pre_uid, post_uid, var in self._out_port_conns:
            if not var in self._out_port_inds_gpot:
                self._out_port_inds_gpot[var] = []
                self._out_var_inds_gpot[var] = []
                self._out_port_inds_spk[var] = []
                self._out_var_inds_spk[var] = []
            ind = self.memory_manager.variables[var]['uids'][pre_uid]
            if post_uid in out_gpot_index:
                # self._out_port_inds_gpot[var].append(self._out_gpot_inds[\
                #                             self._out_gpot_uids.index(post_uid)])
                self._out_port_inds_gpot[var].append(self._out_gpot_inds[\
                                            out_gpot_index[post_uid]])
                self._out_var_inds_gpot[var].append(ind)
            else:
                self._out_port_inds_spk[var].append(self._out_spk_inds[\
                                            out_spk_index[post_uid]])
                self._out_var_inds_spk[var].append(ind)

        tmp = self._out_port_inds_gpot.copy()
        for var in tmp:
            if not self._out_port_inds_gpot[var]:
                del self._out_port_inds_gpot[var]
                del self._out_var_inds_gpot[var]
            else:
                self._out_port_inds_gpot[var] = garray.to_gpu(\
                        np.array(self._out_port_inds_gpot[var],np.int32))
                self._out_var_inds_gpot[var] = garray.to_gpu(\
                        np.array(self._out_var_inds_gpot[var],np.int32))

        tmp = self._out_port_inds_spk.copy()
        for var in tmp:
            if not self._out_port_inds_spk[var]:
                del self._out_port_inds_spk[var]
                del self._out_var_inds_spk[var]
            else:
                self._out_port_inds_spk[var] = garray.to_gpu(\
                        np.array(self._out_port_inds_spk[var],np.int32))
                self._out_var_inds_spk[var] = garray.to_gpu(\
                        np.array(self._out_var_inds_spk[var],np.int32))

    def _setup_input_ports(self):
        self.port_inds_gpot = {}
        self.var_inds_gpot = {}
        self.port_inds_spk = {}
        self.var_inds_spk = {}
        # assuming that the UIDs are unique
        in_gpot_index = {uid: i for i, uid in enumerate(self._in_gpot_uids)}
        in_spk_index = {uid: i for i, uid in enumerate(self._in_spk_uids)}
        for var, uids in self._in_port_vars.items():
            self.port_inds_gpot[var] = []
            self.var_inds_gpot[var] = []
            self.port_inds_spk[var] = []
            self.var_inds_spk[var] = []
            mind = self.memory_manager.variables[var]['models'].index('Port')
            shift = self.memory_manager.variables[var]['cumlen'][mind]
            # The following assumes the intersection of set of variables
            # accessed via spiking with those accessed via gpot ports is null
            for i,uid in enumerate(uids):
                if uid in in_gpot_index:
                    self.port_inds_gpot[var].append(self._in_gpot_inds[\
                                            in_gpot_index[uid]])
                    self.var_inds_gpot[var].append(i + shift)
                else:
                    self.port_inds_spk[var].append(self._in_spk_inds[\
                                            in_spk_index[uid]])
                    self.var_inds_spk[var].append(i + shift)
        tmp = self.port_inds_gpot.copy()
        for var in tmp:
            if not self.port_inds_gpot[var]:
                del self.port_inds_gpot[var]
                del self.var_inds_gpot[var]
            else:
                self.port_inds_gpot[var] = garray.to_gpu(\
                        np.array(self.port_inds_gpot[var],np.int32))
                self.var_inds_gpot[var] = garray.to_gpu(\
                        np.array(self.var_inds_gpot[var],np.int32))
        tmp = self.port_inds_spk.copy()
        for var in tmp:
            if not self.port_inds_spk[var]:
                del self.port_inds_spk[var]
                del self.var_inds_spk[var]
            else:
                self.port_inds_spk[var] = garray.to_gpu(\
                        np.array(self.port_inds_spk[var],np.int32))
                self.var_inds_spk[var] = garray.to_gpu(\
                        np.array(self.var_inds_spk[var],np.int32))


    def init_parameters(self):
        for m, n in self.comp_list:
            if not m in ['Port','Input']:
                nn = n.copy()
                nn.pop(self._uid_key)
                # copy integer and boolean parameters into separate dictionary
                nn_int = {k:v for k, v in iteritems(nn) if (isinstance(v, list)
                            and len(v) and type(v[0]) in [int, bool])}
                nn_rest = {k:v for k, v in iteritems(nn) if (
                           (not isinstance(v, list)) or (len(v) and
                           type(v[0]) not in [int, long, bool]))}
                if nn_int:
                    self.memory_manager.params_htod(m, nn_int, np.int32)
                if nn_rest:
                    self.memory_manager.params_htod(m, nn_rest,
                                                    self._default_dtype)

    def init_variable_memory(self):
        var_info = {}
        for (model, attribs) in self.comp_list:
            if model in ['Port']: continue
            # Add memory for external inputs if required
            if model == 'Input':
                for var, d in iteritems(attribs):
                    if not var in var_info:
                        var_info[var] = {'models':[],'len':[],'delay':0,'uids':[]}
                    var_info[var]['models'].append('Input')
                    var_info[var]['len'].append(len(d[self._uid_key]))
                    var_info[var]['uids'].extend(d[self._uid_key])
                continue
            for var in self._comps[model]['updates']:
                if not var in var_info:
                    var_info[var] = {'models':[],'len':[],'delay':0,'uids':[]}
                var_info[var]['models'].append(model)
                var_info[var]['len'].append(len(attribs[self._uid_key]))
                var_info[var]['uids'].extend(attribs[self._uid_key])

        # Add memory for input ports
        for var in self._in_port_vars:
            if not var in var_info:
                var_info[var] = {'models':[],'len':[],'delay':0,'uids':[]}
            var_info[var]['models'].append('Port')
            var_info[var]['len'].append(len(self._in_port_vars[var]))
            var_info[var]['uids'].extend(self._in_port_vars[var])



        for var in self._variable_delay_map:
            var_info[var]['delay'] = self._variable_delay_map[var]

        for var, d in var_info.items():
            d['cumlen'] = np.cumsum([0]+d['len'])
            d['uids'] = {uid:i for i, uid in enumerate(d['uids'])}

            if isinstance(var, str):
                size = d['cumlen'][-1]
            elif isinstance(var, tuple):
                size = d['cumlen'][-1] * len(var)

            self.memory_manager.memory_alloc(var, size, d['delay']+2,\
                dtype=self._default_dtype,
                info=d)

    def process_connections(self):
        for (model, attribs) in self.comp_list:
            if model in ['Port','Input']: continue
            pre = {var:[] for var in self._comps[model]['accesses']}
            npre = {var:[] for var in self._comps[model]['accesses']}
            data = {var:{} for var in self._comps[model]['accesses']}
            for uid in attribs[self._uid_key]:
                cnt = {var:0 for var in self._comps[model]['accesses']}
                if uid in self.conn_dict:
                    for var in self.conn_dict[uid]:
                        for i in range(len(self.conn_dict[uid][var]['pre'])):
                            # Figure out index of the precomponent in the
                            # particular variable memory
                            p = self.conn_dict[uid][var]['pre'][i]
                            ind = self.memory_manager.variables[var]['uids'][p]
                            pre[var].append(ind)

                            cnt[var] += 1
                            for k in self.conn_dict[uid][var]:
                                if k in ['pre','variable']: continue
                                if k not in data[var]: data[var][k] = []
                                data[var][k].append(self.conn_dict[uid][var][k][i])
                            l = len(pre[var])
                            assert(all([len(data[var][k])==l for k in data[var]]))
                    for var,c in cnt.items():
                        npre[var].append(cnt[var])
                else:
                    for n in npre.values(): n.append(0)
            cumpre = {var: np.cumsum([0]+n) for var, n in npre.items()}

            attribs['pre'] = pre
            attribs['cumpre'] = cumpre
            attribs['npre'] = npre
            attribs['conn_data'] = data

    def post_run(self):
        for comp in self.components.values():
            comp.post_run()
        # Cycle through IO processors as well
        for p in self.input_processors: p.post_run()
        for p in self.output_processors: p.post_run()
        super(LPU, self).post_run()
        if self._print_timing:
            print('time spent on:', self.timing)

    def run_step(self):
        super(LPU, self).run_step()


        # Update input ports
        if self._print_timing:
            start_all = time.time()
            start = time.time()
        self._read_LPU_input()
        if self._print_timing:
            cuda.Context.synchronize()
            self.timing['read_input'] += time.time()-start


        # Fetch updated input if available from all input processors
        if self._print_timing:
            start = time.time()
        for p in self.input_processors: p.run_step()
        if self._print_timing:
            cuda.Context.synchronize()
            self.timing['input_processors'] += time.time()-start

        if self._print_timing:
            start = time.time()
        for model in self._exec_order:
            if model in self._model_var_inj:
                for var in self._model_var_inj[model]:
                    # Reset memory for external input to zero if present
                    self.memory_manager.fill_zeros(model='Input', variable=var)
                    for p in self.input_processors:
                        p.inject_input(var)
        if self._print_timing:
            cuda.Context.synchronize()
            self.timing['inject_input'] += time.time()-start

        # Call run_step of components
        if self._print_timing:
            start = time.time()
        for model in self._exec_order:
            # Get correct position in buffer for update
            update_pointers = {}
            for var in self._comps[model]['updates']:
                buff = self.memory_manager.get_buffer(var)
                mind = self.memory_manager.variables[var]['models'].index(model)
                shift = self.memory_manager.variables[var]['cumlen'][mind]
                buffer_current_plus_one = buff.current + 1
                if buffer_current_plus_one >= buff.buffer_length:
                    buffer_current_plus_one = 0
                update_pointers[var] = int(buff.gpudata)+\
                                       (buffer_current_plus_one*buff.ld+\
                                        shift)*buff.dtype.itemsize
            self.components[model].run_step(update_pointers)
        if self._print_timing:
            cuda.Context.synchronize()
            self.timing['model_run'] += time.time()-start

        # Process output processors
        if self._print_timing:
            start = time.time()
        for p in self.output_processors: p.run_step()
        if self._print_timing:
            cuda.Context.synchronize()
            self.timing['output_processors'] += time.time()-start

        # Check for transforms

        # Update output ports
        if self._print_timing:
            start = time.time()
        self._extract_output()
        if self._print_timing:
            cuda.Context.synchronize()
            self.timing['extract_output'] += time.time()-start

        # Step through buffers
        self.memory_manager.step()

        self.step += 1

        # Instruct Control inteface to process any pending commands
        if self.control_inteface: self.control_inteface.process_commands()
        if self._print_timing:
            cuda.Context.synchronize()
            self.timing['total'] += time.time()-start_all

    def _read_LPU_input(self):
        """
        Extract membrane voltages/spike states from LPU's port map data arrays and
        store them in buffers.
        """
        for var in self.port_inds_gpot:
            # Get correct position in buffer for update
            buff = self.memory_manager.get_buffer(var)
            dest_mem = garray.GPUArray((1,buff.size),buff.dtype,
                                       gpudata=int(buff.gpudata)+\
                                       buff.current*buff.ld*\
                                       buff.dtype.itemsize)
            self.set_inds_both(self.pm['gpot'].data, dest_mem,
                               self.port_inds_gpot[var],self.var_inds_gpot[var])
        for var in self.port_inds_spk:
            # Get correct position in buffer for update
            buff = self.memory_manager.get_buffer(var)
            dest_mem = garray.GPUArray((1,buff.size),buff.dtype,
                                       gpudata=int(buff.gpudata)+\
                                       buff.current*buff.ld*\
                                       buff.dtype.itemsize)
            self.set_inds_both(self.pm['spike'].data, dest_mem, \
                          self.port_inds_spk[var],self.var_inds_spk[var])

    def _extract_output(self):
        """
        Extract membrane voltages/spike states from LPU's port map data arrays and
        store them in buffers.
        """

        for var in self._out_port_inds_gpot:
            # Get correct position in buffer for update
            buff = self.memory_manager.get_buffer(var)
            src_mem = garray.GPUArray((1,buff.size),buff.dtype,
                                       gpudata=int(buff.gpudata)+\
                                       buff.current*buff.ld*\
                                      buff.dtype.itemsize)
            self.set_inds_both(src_mem, self.pm['gpot'].data, \
                    self._out_var_inds_gpot[var], self._out_port_inds_gpot[var])
        for var in self._out_port_inds_spk:
            # Get correct position in buffer for update
            buff = self.memory_manager.get_buffer(var)
            src_mem = garray.GPUArray((1,buff.size),buff.dtype,
                                       gpudata=int(buff.gpudata)+\
                                       buff.current*buff.ld*\
                                       buff.dtype.itemsize)
            self.set_inds_both(src_mem, self.pm['spike'].data, \
                    self._out_var_inds_spk[var], self._out_port_inds_spk[var])

    def set_inds_both(self, src, dest, src_inds, dest_inds):
        """
        Set `dest[dest_inds[i]] = src[src_inds[i]] for i in range(len(src_inds))`
        """

        try:
            func = self.set_inds_both.cache[(src_inds.dtype, src.dtype)]
        except KeyError:
            inds_ctype = dtype_to_ctype(src_inds.dtype)
            data_ctype = dtype_to_ctype(src.dtype)
            v = ("{data_ctype} *dest, {inds_ctype} *dest_inds, " +\
                 "{inds_ctype} *src_inds, {data_ctype} *src").format(\
                        data_ctype=data_ctype,inds_ctype=inds_ctype)
            func = elementwise.ElementwiseKernel(v,\
                            "dest[dest_inds[i]] = src[src_inds[i]]")
            self.set_inds_both.cache[(src_inds.dtype, src.dtype)] = func
        func(dest, dest_inds, src_inds, src, range=slice(0, len(src_inds), 1) )

    set_inds_both.cache = {}

    def _instantiate_component(self, comp_name):
        try:
            cls = self._comps[comp_name]['cls']
        except:
            self.log_info("Error instantiating component of model '%s'" \
                          % comp_name)
            return None

        params_dict = self.memory_manager.parameters[comp_name]
        access_buffers = {var:self.memory_manager.get_buffer(var) \
                          for var in self._comps[comp_name]['accesses'] \
                          if var in self.memory_manager.variables}
        return cls(params_dict, access_buffers, self.dt,
                   LPU_id=self.LPU_id, debug=self._debug,
                   cuda_verbose=bool(self._compile_options))


    def _load_components(self, extra_comps=[]):
        """
        Load all available NDcomponents
        """
        child_classes = NDComponent.NDComponent.__subclasses__()
        comp_classes = child_classes[:]
        for cls in child_classes:
            comp_classes.extend(cls.__subclasses__())
        comp_classes.extend(extra_comps)
        self._comps = {cls.__name__:{'accesses': cls.accesses ,
                                     'updates':cls.updates,
                                     'cls':cls} \
                       for cls in comp_classes if not cls.__name__[:4]=='Base'}

    @property
    def time(self):
        return self.step*self.dt

class uid_generator(object):
    def __init__(self):
        self.input_count = 0
        self.auto_count = 0

    def generate_uid(self, input=False):
        if input:
            uid = 'input_' + str(self.input_count)
            self.input_count += 1
        else:
            uid = 'auto_' + str(self.auto_count)
            self.auto_count += 1
        return uid



import numpy as np

from .BaseInputProcessor import BaseInputProcessor


class StepInputProcessor(BaseInputProcessor):

    def __init__(self, variable, uids, val, start, stop,
                 input_file = None, input_interval = 1,
                 sensory_file = None, sensory_interval = 1):
        super(StepInputProcessor, self).__init__([(variable, uids)],
                                                 mode = 1,
                                                 memory_mode = 'gpu',
                                                 input_file = input_file,
                                                 input_interval = input_interval,
                                                 sensory_file = sensory_file,
                                                 sensory_interval = sensory_interval)
        self.start = start
        self.stop = stop
        self.var = variable
        self.num = len(uids)
        self.started = False
        self.stopped = False

        if np.isscalar(val):
            self.val = np.full((self.num,), val)
        else:
            assert len(val) == self.num, \
                f"Step Input specified with {self.num} uids but got input value of length {len(val)}"
            self.val = val.copy()


    def update_input(self):
        if self.stopped:
            self.variables[self.var]['input'].fill(0)
        else:
            if self.started:
                self.variables[self.var]['input'].set(self.val)
        # if self.LPU_obj.time == self.start:
        #     self.variables[self.var]['input'].fill(self.val) # * np.ones(self.num, self.dtypes[self.var])
        # else:
        #     self.variables[self.var]['input'].fill(0)

    def is_input_available(self):
        if not self.started:
            if self.LPU_obj.time >= self.start:
                self.started = True
                return True
        else:
            if not self.stopped:
                if self.LPU_obj.time >= self.stop:
                    self.stopped = True
                    return True
        return False

    def post_run(self):
        super(StepInputProcessor, self).post_run()




from future.utils import listvalues
import numpy as np
import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype, context_dependent_memoize
# import pycuda.elementwise as elementwise
from pycuda.compiler import SourceModule
import pycuda.driver as cuda

from neurokernel.LPU.LPU import LPU

class BaseOutputProcessor(object):
    def __init__(self, var_list, sample_interval = 1):
        # var_list should be a list of (variable, uids)
        # Invalid uids will be ignored
        # if uids is None, the entire variable will be outputted
        # If uids are specified, the order will not be respected.
        # Derived classes should access self.variables[variable]['uids']
        # at pre_run to know the order
        # Output will be stored in self.variables[variable]['output']
        # and should be processed by derived classes in process_output
        self.variables = {var:{'uids':uids,'output':None}
                          for var, uids in var_list}
        self.sample_interval = sample_interval
        self.epoch = 0
        self.src_inds = {}
        self._LPU_obj = None
        self._d_output = {}

    @property
    def LPU_obj(self):
        return self._LPU_obj

    @LPU_obj.setter
    def LPU_obj(self, value):
        assert(isinstance(value, LPU))
        self._LPU_obj = value
        self.start_time = self._LPU_obj.time
        self.dt = self._LPU_obj.dt
        self.sim_dt = self._LPU_obj.dt
        self.memory_manager = self._LPU_obj.memory_manager

    def run_step(self):
        assert(self.LPU_obj)
        self.epoch += 1
        if self.epoch == self.sample_interval:
            self.epoch = 0
            for var, d in self.variables.items():
                buff = self.memory_manager.get_buffer(var)
                # src_mem = garray.GPUArray((1,buff.size),buff.dtype,
                #                           gpudata=int(buff.gpudata)+\
                #                           buff.current*buff.ld*\
                #                           buff.dtype.itemsize)
                src_mem = int(buff.gpudata)+buff.current*buff.ld*buff.dtype.itemsize
                dest_mem = self.get_output_array(var)
                if dest_mem is None:
                    dest_mem = self._d_output[var].gpudata
                self.get_inds(var, src_mem, dest_mem)
            self.process_output()

        if 'spike_state' in self.variables:
            var = 'spike_state'
            buff = self.memory_manager.get_buffer(var)
            src_mem = int(buff.gpudata)+buff.current*buff.ld*buff.dtype.itemsize
            dest_mem = self.get_output_array(var)
            if dest_mem is None:
                dest_mem = self._d_output[var].gpudata
            self.get_inds(var, src_mem, dest_mem)
            self.process_spike_output()

    def get_output_array(self, var):
        return None

    def _pre_run(self):
        assert(self.LPU_obj)
        assert(all([var in self.memory_manager.variables
                    for var in self.variables]))
        self.get_inds_func = {}
        for var, d in self.variables.items():
            v_dict =  self.memory_manager.variables[var]
            if not d['uids']:
                uids = list(v_dict['uids'])
                inds = listvalues(v_dict['uids'])
                o = np.argsort(inds)
                d['uids'] = [uids[i] for i in o]
                self.src_inds[var] = garray.to_gpu(np.arange(len(d['uids'])))
            else:
                uids = []
                inds = []
                for uid in d['uids']:
                    try:
                        inds.append(v_dict['uids'][uid])
                        uids.append(uid)
                    except:
                        pass
                inds = np.array(inds,np.int32)
                o = np.argsort(inds)
                self.src_inds[var] = garray.to_gpu(inds[o])
                d['uids'] = [uids[i] for i in o]
            self._d_output[var] = garray.empty(len(d['uids']),
                                               v_dict['buffer'].dtype)
            d['output']=np.zeros(len(d['uids']), v_dict['buffer'].dtype)
            self.get_inds_func[var] = get_inds_kernel(self.src_inds[var].dtype,
                                                      v_dict['buffer'].dtype)
        self.pre_run()

    def get_output_gpu(self, var):
        return self._d_output[var]

    def get_output(self, var):
        self._d_output[var].get(self.variables[var]['output'])
        return self.variables[var]['output']

    # Should be implemented by child class
    def pre_run(self):
        pass

    # Should be implemented by child class
    def process_output(self):
        pass

    def process_spike_output(self):
        pass

    # Should be implemented by child class
    def post_run(self):
        pass

    def get_inds(self, var, src, dest, src_shift = 0):
        func = self.get_inds_func[var]
        inds = self.src_inds[var]
        func.prepared_async_call(
            func.grid, func.block, None,
            dest, int(src_shift), inds.gpudata, src, inds.size)

#     def get_inds1(self, src_dtype, src, dest, inds, src_shift = 0):
#         # assert src.dtype == dest.dtype
#         inds_ctype = dtype_to_ctype(inds.dtype)
#         data_ctype = dtype_to_ctype(src_dtype)
#
#         func = get_inds_kernel1(inds_ctype, data_ctype)
#         func.prepared_async_call(
#             func.grid, func.block, None,
#             dest, int(src_shift), inds.gpudata, src, inds.size)
#
#     def get_inds(self, src, dest, inds, src_shift=0):
#         """
#         Set `dest[i] = src[src_shift+inds[i]] for i in range(len(inds))`
#         """
#
#         assert src.dtype == dest.dtype
#         inds_ctype = dtype_to_ctype(inds.dtype)
#         data_ctype = dtype_to_ctype(src.dtype)
#
#         func = get_inds_kernel(inds_ctype, data_ctype)
#         func(dest, int(src_shift), inds, src, range=slice(0, len(inds), 1) )
#
# @context_dependent_memoize
# def get_inds_kernel(inds_ctype, src_ctype):
#     v = ("{data_ctype} *dest, int src_shift, " +\
#          "{inds_ctype} *inds, {data_ctype} *src").format(\
#                 data_ctype=src_ctype,inds_ctype=inds_ctype)
#     func = elementwise.ElementwiseKernel(v,\
#                     "dest[i] = src[src_shift+inds[i]]")
#     return func


@context_dependent_memoize
def get_inds_kernel(inds_dtype, src_dtype):
    template = """
__global__ void update(%(data_ctype)s* dest, int src_shift,
                       %(inds_ctype)s* inds, %(data_ctype)s* src, int N)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    for(int i = tid; i < N; i += total_threads)
    {
        dest[i] = src[src_shift+inds[i]];
    }
}
"""
    mod = SourceModule(template % {"data_ctype": dtype_to_ctype(src_dtype),
                                   "inds_ctype": dtype_to_ctype(inds_dtype)})
    func = mod.get_function("update")
    func.prepare('PiPPi')
    func.block = (128,1,1)
    func.grid = (16 * cuda.Context.get_device().MULTIPROCESSOR_COUNT, 1)
    return func



#!/usr/bin/env python

from collections import OrderedDict

import numpy as np

import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from abc import ABCMeta, abstractmethod, abstractproperty
from future.utils import with_metaclass

from neurokernel.LPU.NDComponents.NDComponent import NDComponent

class BaseSynapseModel(with_metaclass(ABCMeta, NDComponent)):
    # __metaclass__ = ABCMeta

    accesses = ['V']
    updates = ['g']
    params = []
    extra_params = []
    internals = []

    def __init__(self, params_dict, access_buffers, dt,
                 LPU_id=None, debug=False, cuda_verbose=False):
        if cuda_verbose:
            self.compile_options = ['--ptxas-options=-v']
        else:
            self.compile_options = []

        self.debug = debug
        self.num_comps = params_dict[self.params[0]].size
        self.dtype = params_dict[self.params[0]].dtype
        self.LPU_id = LPU_id
        self.dt = np.double(dt)
        #self.ddt = np.double(1e-6)
        #self.steps = np.int32(max( int(self.dt/self.ddt), 1 ))

        self.params_dict = params_dict
        self.access_buffers = access_buffers

        self.internal_states = {
            c: garray.zeros(self.num_comps, dtype = self.dtype)+self.internals[c] \
            for c in self.internals}

        self.inputs = {
            k: garray.empty(self.num_comps, dtype = self.access_buffers[k].dtype)\
            for k in self.accesses}

        self.retrieve_buffer_funcs = {}
        for k in self.accesses:
            self.retrieve_buffer_funcs[k] = \
                self.get_retrieve_buffer_func(
                    k, dtype = self.access_buffers[k].dtype)

        dtypes = {'dt': self.dtype}
        dtypes.update({'input_{}'.format(k): self.inputs[k].dtype for k in self.accesses})
        dtypes.update({'param_{}'.format(k): self.params_dict[k].dtype for k in self.params})
        dtypes.update({'internal_{}'.format(k): self.internal_states[k].dtype for k in self.internals})
        dtypes.update({'update_{}'.format(k): self.dtype for k in self.updates})
        self.update_func = self.get_update_func(dtypes)

    def retrieve_buffer(self, param, st = None):
        self.retrieve_buffer_funcs[param].prepared_async_call(
            self.retrieve_buffer_funcs[param].grid,
            self.retrieve_buffer_funcs[param].block,
            st,
            self.access_buffers[param].gpudata,
            self.access_buffers[param].ld,                    #i
            self.access_buffers[param].current,               #i
            self.access_buffers[param].buffer_length,         #i
            self.params_dict['pre'][param].gpudata,                #P
            self.params_dict['npre'][param].gpudata,               #P
            self.params_dict['cumpre'][param].gpudata,             #P
            self.params_dict['conn_data'][param]['delay'].gpudata,
            self.inputs[param].gpudata,
            self.num_comps)

    def get_retrieve_buffer_func(self, param, dtype):
        template = """
__global__ void retrieve(%(type)s* buffer, int buffer_ld, int current,
                         int buffer_length, int* pre, int* npre,
                         int* cumpre, int* delay,
                         %(type)s* linearized_buffer, int n_items)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    int dl, col;
    for(int i = tid; i < n_items; i += total_threads)
    {
        // npre[i] here is always 1, so no else is provided.
        if(npre[i])
        {
            dl = delay[i];
            col = current - dl;
            if(col < 0)
            {
                col = buffer_length + col;
            }
            linearized_buffer[i] = buffer[col*buffer_ld + pre[cumpre[i]]];
        }
    }
}
        """
        mod = SourceModule(template % {"type":dtype_to_ctype(dtype)},
                           options=self.compile_options)
        func = mod.get_function("retrieve")
        func.prepare('PiiiPPPPPi')
        func.block = (256,1,1)
        func.grid = (min(6 * cuda.Context.get_device().MULTIPROCESSOR_COUNT,
                         (self.num_comps-1) // 256 + 1), 1)
        return func

    def run_step(self, update_pointers, st = None):
        # retrieve all buffers into a linear array
        for k in self.inputs:
            self.retrieve_buffer(k, st = st)

        self.update_func.prepared_async_call(
            self.update_func.grid, self.update_func.block, st,
            self.num_comps, self.internal_dt, self.internal_steps,
            *[self.inputs[k].gpudata for k in self.accesses]+\
            [self.params_dict[k].gpudata for k in self.params]+\
            [self.internal_states[k].gpudata for k in self.internals]+\
            [update_pointers[k] for k in self.updates])

    def get_update_func(self, dtypes):
        type_dict = {k: dtype_to_ctype(dtypes[k]) for k in dtypes}
        type_dict.update({'fletter': 'f' if type_dict['param_{}'.format(self.params[0])] == 'float' else ''})
        mod = SourceModule(self.get_update_template() % type_dict,
                           options=self.compile_options)
        func = mod.get_function("update")
        func.prepare('i'+np.dtype(dtypes['dt']).char+'i'+'P'*(len(type_dict)-2))
        func.block = (256,1,1)
        func.grid = (min(6 * cuda.Context.get_device().MULTIPROCESSOR_COUNT,
                         (self.num_comps-1) // 256 + 1), 1)
        return func



from collections import OrderedDict


import numpy as np
import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from .SynapseAMPA import SynapseAMPA

# assumes a maximum of one input connection per synapse
class SynapseGABA(SynapseAMPA):
    pass

if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.FileInputProcessor import FileInputProcessor
    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('synapse0', {
               'class': 'SynapseGABA',
               'name': 'SynapseGABA',
               'gmax': 0.0021788,
               'st': 100.0,
               'reverse': -80.0
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    #fl_input_processor = StepInputProcessor('I', ['neuron0'], 1, 0.2, 0.4)
    fl_input_processor = FileInputProcessor('input_spike.h5')
    fl_output_processor = FileOutputProcessor([('g', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()



try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)

# Ignore all exceptions so that this doesn't cause package installation
# to fail if pkg_resources can't find neurokernel:
try:
    from version import __version__
except:
    pass




import time

import numpy as np
import pycuda.driver as cuda
import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype, context_dependent_memoize
from pycuda.compiler import SourceModule

from .BaseInputProcessor import BaseInputProcessor
from ..utils import curand

class BernoulliInputProcessor(BaseInputProcessor):

    def __init__(self, variable, uids, p, start, stop, seed = None):
        super(BernoulliInputProcessor, self).__init__([(variable, uids)],
                                                      mode = 0,
                                                      memory_mode = 'gpu')
        self.p = p
        self.start = start
        self.stop = stop
        self.var = variable
        self.seed = seed
        self.num = len(uids)

    def update_input(self):
        #self.variables[self.var]['input'] = np.random.binomial(1, self.p, size=(self.num,)).astype(self.dtypes[self.var])
        self.update_kernel.prepared_async_call(
            self.update_kernel.grid, self.update_kernel.block, None,
            self.variables[self.var]['input'].gpudata, self.num,
            self.p, self.state.gpudata)

    def is_input_available(self):
        return (self.LPU_obj.time >= self.start and
                self.LPU_obj.time < self.stop)

    def pre_run(self):
        self.update_kernel = get_update_func(self.dtypes[self.var])
        self.state = curand.curand_setup(
            self.update_kernel.block[0]*self.update_kernel.grid[0],
            self.seed if self.seed is not None else time.monotonic_ns())

    def post_run(self):
        pass


@context_dependent_memoize
def get_update_func(dtype):
    template = """
#include "curand_kernel.h"
extern "C" {
__global__ void draw(%(type)s* output, int N, double p, curandStateXORWOW_t* state)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    curandStateXORWOW_t local_state = state[tid];
    double n;
    for(int i = tid; i < N; i += total_threads)
    {
        n = curand_uniform_double(&local_state);
        output[i] = (n < p) ? 1 : 0;
    }
    state[tid] = local_state;
}
}
"""
    mod = SourceModule(template % {"type": dtype_to_ctype(dtype)},
                       no_extern_c = True)
    func = mod.get_function("draw")
    func.prepare('PidP')
    func.block = (128,1,1)
    func.grid = (16 * cuda.Context.get_device().MULTIPROCESSOR_COUNT, 1)
    return func



from collections import OrderedDict


import numpy as np
import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from .BaseSynapseModel import BaseSynapseModel

# assumes a maximum of one input connection per synapse
class SynapseAMPA(BaseSynapseModel):
    accesses = ['spike_state']
    updates = ['g']
    params = ['gmax', 'st']
    internals = OrderedDict([('tst', 0.)])

    @property
    def maximum_dt_allowed(self):
        return 1e-4

    def get_update_template(self):
        template = """
__global__ void update(int num_comps, %(dt)s dt, int nsteps,
                       %(input_spike_state)s* g_spike_state,
                       %(param_gmax)s *g_gmax, %(param_st)s *g_st,
                       %(internal_tst)s *g_tst,
                       %(update_g)s *g_g)
{
    int tid = threadIdx.x + blockIdx.x*blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(dt)s ddt = dt*1000.;
    %(input_spike_state)s spike_state;
    %(param_st)s st, est;
    %(internal_tst)s tst;

    for(int i = tid; i < num_comps; i += total_threads)
    {
        spike_state = g_spike_state[i];
        st = g_st[i];
        tst = g_tst[i];

        est = exp%(fletter)s(-ddt/st);

        for(int k = 0; k < nsteps; ++k)
        {
            tst *= est;
            if(k == 0 && (spike_state>0.0))
            {
                tst += 1;
            }
        }

        g_g[i] = tst*g_gmax[i];
        g_tst[i] = tst;
    }
}
"""
        return template


if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.FileInputProcessor import FileInputProcessor
    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('synapse0', {
               'class': 'SynapseAMPA',
               'name': 'SynapseAMPA',
               'gmax': 0.003,#21.788,
               'st': 20.0,#100.0,
               'reverse': 0.0
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    #fl_input_processor = StepInputProcessor('I', ['neuron0'], 1, 0.2, 0.4)
    fl_input_processor = FileInputProcessor('input_spike.h5')
    fl_output_processor = FileOutputProcessor([('g', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()
